from __future__ import annotations

__all__ = ["SequentialInitializer"]

import logging
from collections.abc import Sequence

from coola.utils import str_indent, str_sequence
from torch.nn import Module

from gravitorch.nn.init.base import BaseInitializer
from gravitorch.nn.init.factory import setup_initializer

logger = logging.getLogger(__name__)


class SequentialInitializer(BaseInitializer):
    r"""Implements a module initializer that sequentially calls module
    initializers.

    Args:
    ----
        initializers: Specifies the sequence of module initializers.
            The sequence order defines the order of the call.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.nn.init import ConstantBias, KaimingNormal, SequentialInitializer
        >>> from torch import nn
        >>> module = nn.Sequential(nn.Linear(4, 6), nn.ReLU(), nn.BatchNorm1d(6), nn.Linear(6, 1))
        >>> initializer = SequentialInitializer([KaimingNormal(), ConstantBias(value=0.0)])
        >>> initializer
        SequentialInitializer(
          (0): KaimingNormal(mode=fan_in, nonlinearity=leaky_relu, neg_slope=0.0, learnable_only=True, log_info=False)
          (1): ConstantBias(value=0.0, learnable_only=True, log_info=False)
        )
        >>> initializer.initialize(module)
    """

    def __init__(
        self,
        initializers: Sequence[BaseInitializer | dict],
    ) -> None:
        super().__init__()
        self._initializers = tuple(setup_initializer(initializer) for initializer in initializers)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  {str_indent(str_sequence(self._initializers))}\n)"
        )

    def initialize(self, module: Module) -> None:
        for i, initializer in enumerate(self._initializers):
            logger.info(f"[{i}/{len(self._initializers)}] {initializer}")
            initializer.initialize(module)



r"""This module defines the absolute error metric."""

from __future__ import annotations

__all__ = ["AbsoluteError"]

import logging

from torch import Tensor

from gravitorch.models.metrics.base_epoch import BaseStateEpochMetric
from gravitorch.models.metrics.state import BaseState, ErrorState
from gravitorch.nn.functional import absolute_error

logger = logging.getLogger(__name__)


class AbsoluteError(BaseStateEpochMetric):
    r"""Implements the absolute error metric.

    Args:
    ----
        mode (str): Specifies the mode (e.g ``'train'`` or ``'eval'``).
        name (str, optional): Specifies the name of the metric. The
            name is used to log the metric results.
            Default: ``'abs_err'``
        state (``BaseState`` or dict, optional): Specifies the metric
            state or its configuration. If ``None``, ``ErrorState`` is
            instantiated. Default: ``None``

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from gravitorch.models.metrics import AbsoluteError
        >>> metric = AbsoluteError("eval")
        >>> metric
        AbsoluteError(
          (mode): eval
          (name): abs_err
          (state): ErrorState(num_predictions=0)
        )
        >>> metric(torch.ones(2, 4), torch.ones(2, 4))
        >>> metric.value()
        {'eval/abs_err_mean': 0.0,
         'eval/abs_err_min': 0.0,
         'eval/abs_err_max': 0.0,
         'eval/abs_err_sum': 0.0,
         'eval/abs_err_num_predictions': 8}
        >>> metric(torch.eye(2), torch.ones(2, 2))
        >>> metric.value()
        {'eval/abs_err_mean': 0.16666666666666666,
         'eval/abs_err_min': 0.0,
         'eval/abs_err_max': 1.0,
         'eval/abs_err_sum': 2.0,
         'eval/abs_err_num_predictions': 12}
        >>> metric.reset()
        >>> metric(torch.eye(2), torch.ones(2, 2))
        >>> metric.value()
        {'eval/abs_err_mean': 0.5,
         'eval/abs_err_min': 0.0,
         'eval/abs_err_max': 1.0,
         'eval/abs_err_sum': 2.0,
         'eval/abs_err_num_predictions': 4}
    """

    def __init__(
        self,
        mode: str,
        name: str = "abs_err",
        state: BaseState | dict | None = None,
    ) -> None:
        super().__init__(mode=mode, name=name, state=state or ErrorState())

    def forward(self, prediction: Tensor, target: Tensor) -> None:
        r"""Updates the mean absolute error metric given a mini-batch of
        examples.

        Args:
        ----
            prediction (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn)`` and type float or long):
                Specifies the predictions.
            target (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn)`` and type float or long):
                Specifies the target tensor.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from gravitorch.models.metrics import AbsoluteError
            >>> metric = AbsoluteError("eval")
            >>> metric(torch.ones(2, 4), torch.ones(2, 4))
            >>> metric.value()
            {'eval/abs_err_mean': 0.0,
             'eval/abs_err_min': 0.0,
             'eval/abs_err_max': 0.0,
             'eval/abs_err_sum': 0.0,
             'eval/abs_err_num_predictions': 8}
        """
        self._state.update(absolute_error(prediction, target))



from __future__ import annotations

__all__ = ["DictOfListConverterIterDataPipe", "ListOfDictConverterIterDataPipe"]

from collections.abc import Hashable, Iterator, Mapping, Sequence

from coola.utils import str_indent, str_mapping
from torch.utils.data import IterDataPipe

from gravitorch.utils.mapping import convert_to_dict_of_lists, convert_to_list_of_dicts


class DictOfListConverterIterDataPipe(IterDataPipe[dict[Hashable, list]]):
    r"""Implements an ``IterDataPipe`` to convert a sequence of mappings
    to a dictionary of lists.

    Args:
    ----
        datapipe (``IterDataPipe``): Specifies an ``IterDataPipe``
            of sequences of mappings.

    Example usage:

    .. code-block:: pycon

        >>> from torch.utils.data.datapipes.iter import IterableWrapper
        >>> from gravitorch.datapipes.iter import DictOfListConverter
        >>> dp = DictOfListConverter(
        ...     IterableWrapper(
        ...         [
        ...             [{"key1": 1, "key2": 10}, {"key1": 2, "key2": 20}, {"key1": 3, "key2": 30}],
        ...             [{"key": "a"}, {"key": -2}],
        ...         ]
        ...     ),
        ... )
        >>> dp
        DictOfListConverterIterDataPipe(
          (datapipe): IterableWrapperIterDataPipe
        )
        >>> list(dp)
        [{'key1': [1, 2, 3], 'key2': [10, 20, 30]}, {'key': ['a', -2]}]
    """

    def __init__(self, datapipe: IterDataPipe[Sequence[Mapping]]) -> None:
        self._datapipe = datapipe

    def __iter__(self) -> Iterator[dict[Hashable, list]]:
        for data in self._datapipe:
            yield convert_to_dict_of_lists(data)

    def __len__(self) -> int:
        return len(self._datapipe)

    def __repr__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        args = str_indent(str_mapping({"datapipe": self._datapipe}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"


class ListOfDictConverterIterDataPipe(IterDataPipe[list[dict]]):
    r"""Implements an ``IterDataPipe`` to convert a mapping of sequences
    to a list of dictionaries.

    Args:
    ----
        datapipe (``IterDataPipe``): Specifies an
            ``IterDataPipe`` of mappings of sequences.

    Example usage:

    .. code-block:: pycon

        >>> from torch.utils.data.datapipes.iter import IterableWrapper
        >>> from gravitorch.datapipes.iter import ListOfDictConverter
        >>> dp = ListOfDictConverter(
        ...     IterableWrapper([{"key1": [1, 2, 3], "key2": [10, 20, 30]}, {"key": ["a", -2]}]),
        ... )
        >>> dp
        ListOfDictConverterIterDataPipe(
          (datapipe): IterableWrapperIterDataPipe
        )
        >>> list(dp)
        [[{'key1': 1, 'key2': 10}, {'key1': 2, 'key2': 20}, {'key1': 3, 'key2': 30}], [{'key': 'a'}, {'key': -2}]]
    """

    def __init__(self, datapipe: IterDataPipe[Mapping[Hashable, Sequence]]) -> None:
        self._datapipe = datapipe

    def __iter__(self) -> Iterator[list[dict]]:
        for data in self._datapipe:
            yield convert_to_list_of_dicts(data)

    def __len__(self) -> int:
        return len(self._datapipe)

    def __repr__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        args = str_indent(str_mapping({"datapipe": self._datapipe}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"



from __future__ import annotations

__all__ = ["BaseDataCreator", "is_datacreator_config", "setup_datacreator"]

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Generic, TypeVar

from objectory import AbstractFactory
from objectory.utils import is_object_config

from gravitorch.utils.format import str_target_object

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine

logger = logging.getLogger(__name__)

T = TypeVar("T")


class BaseDataCreator(ABC, Generic[T], metaclass=AbstractFactory):
    r"""Defines the base class to implement a data creator.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.data.datacreators import HypercubeVertexDataCreator
        >>> creator = HypercubeVertexDataCreator(num_examples=10, num_classes=5, feature_size=6)
        >>> creator
        HypercubeVertexDataCreator(num_examples=10, num_classes=5, feature_size=6, noise_std=0.2, random_seed=15782179921860610490)
        >>> data = creator.create()
        >>> data
        {'target': tensor([...]), 'input': tensor([[...]])}
    """

    @abstractmethod
    def create(self, engine: BaseEngine | None = None) -> T:
        r"""Creates data.

        Args:
        ----
            engine (``BaseEngine`` or ``None``): Specifies an engine.
                Default: ``None``

        Returns:
        -------
            The created data.

        Example usage:

        .. code-block:: pycon

            >>> from gravitorch.data.datacreators import HypercubeVertexDataCreator
            >>> creator = HypercubeVertexDataCreator(num_examples=10, num_classes=5, feature_size=6)
            >>> data = creator.create()
            >>> data
            {'target': tensor([...]), 'input': tensor([[...]])}
        """


def is_datacreator_config(config: dict) -> bool:
    r"""Indicates if the input configuration is a configuration for a
    ``BaseDataCreator``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
    ----
        config (dict): Specifies the configuration to check.

    Returns:
    -------
        bool: ``True`` if the input configuration is a configuration
            for a ``BaseDataCreator`` object.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.data.datacreators import is_datacreator_config
        >>> is_datacreator_config(
        ...     {"_target_": "gravitorch.data.datacreators.HypercubeVertexDataCreator"}
        ... )
        True
    """
    return is_object_config(config, BaseDataCreator)


def setup_datacreator(data_creator: BaseDataCreator | dict) -> BaseDataCreator:
    r"""Sets up a data creator.

    The data creator is instantiated from its configuration by using
    the ``BaseDataCreator`` factory function.

    Args:
    ----
        data_creator (``BaseDataCreator`` or dict): Specifies the data
            creator or its configuration.

    Returns:
    -------
        ``BaseDataCreator``: The instantiated data creator.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.data.datacreators import setup_datacreator
        >>> datacreator = setup_datacreator(
        ...     {
        ...         "_target_": "gravitorch.data.datacreators.HypercubeVertexDataCreator",
        ...         "num_examples": 10,
        ...         "num_classes": 5,
        ...         "feature_size": 6,
        ...     }
        ... )
        >>> datacreator
        HypercubeVertexDataCreator(num_examples=10, num_classes=5, feature_size=6, noise_std=0.2, random_seed=15782179921860610490)
    """
    if isinstance(data_creator, dict):
        logger.info(
            "Initializing a data creator from its configuration... "
            f"{str_target_object(data_creator)}"
        )
        data_creator = BaseDataCreator.factory(**data_creator)
    return data_creator



r"""This module defines the base class for the data loader collators."""

from __future__ import annotations

__all__ = ["BaseCollator", "setup_collator"]

from abc import ABC, abstractmethod
from collections.abc import Callable
from typing import Generic, TypeVar

from objectory import AbstractFactory
from torch.utils.data.dataloader import default_collate

R = TypeVar("R")
T = TypeVar("T")


class BaseCollator(Generic[T, R], Callable[[list[T]], R], ABC, metaclass=AbstractFactory):
    r"""Defines the base class to create a batch of examples.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from gravitorch.dataloaders.collators import DictPaddedSequenceCollator
        >>> collator = DictPaddedSequenceCollator(keys_to_pad=["feature"])
        >>> collator
        DictPaddedSequenceCollator(keys_to_pack=('feature',), batch_first=False, padding_value=0.0)
        >>> data = [
        ...     {"feature": torch.full((2,), 2.0)},
        ...     {"feature": torch.full((3,), 3.0)},
        ...     {"feature": torch.full((4,), 4.0)},
        ... ]
        >>> batch = collator(data)
        >>> batch
        {'feature': tensor([[4., 3., 2.],
                [4., 3., 2.],
                [4., 3., 0.],
                [4., 0., 0.]])}
    """

    @abstractmethod
    def __call__(self, data: list[T]) -> R:
        r"""Creates a batch given a list of examples.

        Args:
        ----
            data (list): Specifies a list of examples.

        Returns:
        -------
             A batch of examples.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from gravitorch.dataloaders.collators import DictPaddedSequenceCollator
            >>> collator = DictPaddedSequenceCollator(keys_to_pad=["feature"])
            >>> data = [
            ...     {"feature": torch.full((2,), 2.0)},
            ...     {"feature": torch.full((3,), 3.0)},
            ...     {"feature": torch.full((4,), 4.0)},
            ... ]
            >>> batch = collator(data)
            >>> batch
            {'feature': tensor([[4., 3., 2.],
                    [4., 3., 2.],
                    [4., 3., 0.],
                    [4., 0., 0.]])}
        """


def setup_collator(collator: Callable[[list[T]], R] | dict | None) -> Callable[[list[T]], R]:
    r"""Sets up a data loader collator.

    Args:
    ----
        collator (``Callable`` or dict or None): Specifies the
            data loader collator or its configuration. If ``None``,
            the default data loader collator is used.

    Returns:
    -------
        ``Callable``: The data loader collator.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.dataloaders.collators import setup_collator
        >>> collator = setup_collator(
        ...     {"_target_": "gravitorch.dataloaders.collators.PaddedSequenceCollator"}
        ... )
        >>> collator
        PaddedSequenceCollator(length_key=length, batch_first=False, padding_value=0.0)
    """
    if collator is None:
        collator = default_collate
    if isinstance(collator, dict):
        collator = BaseCollator.factory(**collator)
    return collator



from __future__ import annotations

__all__ = ["TrainingRunner"]

import logging
from collections.abc import Sequence

from coola.utils import str_indent, str_mapping, str_sequence

from gravitorch.distributed import comm as dist
from gravitorch.engines.base import BaseEngine
from gravitorch.handlers import setup_and_attach_handlers
from gravitorch.handlers.base import BaseHandler
from gravitorch.rsrc.base import BaseResource
from gravitorch.runners.resource import BaseResourceRunner
from gravitorch.utils.exp_trackers import setup_exp_tracker
from gravitorch.utils.exp_trackers.base import BaseExpTracker
from gravitorch.utils.seed import manual_seed

logger = logging.getLogger(__name__)


class TrainingRunner(BaseResourceRunner):
    r"""Implements a runner to train a ML model.

    Internally, this runner does the following steps:

        - set the experiment tracker
        - set the random seed
        - instantiate the engine
        - set up and attach the handlers
        - train the model with the engine

    Args:
    ----
        engine (``BaseEngine`` or dict): Specifies the engine or its
            configuration.
        handlers (list or tuple or ``None``): Specifies the list of
            handlers or their configuration. The handlers will be
            attached to the engine. If ``None``, no handler is
            attached to the engine. Default: ``None``
        exp_tracker (``BaseExpTracker`` or dict or None): Specifies
            the experiment tracker or its configuration. If ``None``,
            the no-operation experiment tracker is used.
        random_seed (int, optional): Specifies the random seed.
            Default: ``10139531598155730726``
        resources (sequence or ``None``, optional): Specifies a
            sequence of resources or their configurations.
            Default: ``None``

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.runners import TrainingRunner
        >>> from gravitorch.testing import create_dummy_engine
        >>> engine = create_dummy_engine()
        >>> runner = TrainingRunner(engine)
        >>> runner  # doctest:+ELLIPSIS
        TrainingRunner(
          (engine): AlphaEngine(
              (datasource): DummyDataSource(
                  (datasets):
                    (train): DummyDataset(num_examples=4, feature_size=4)
                    (eval): DummyDataset(num_examples=4, feature_size=4)
                  (dataloader_creators):
                    (train): DataLoaderCreator(
                        (seed): 0
                        (batch_size): 2
                        (shuffle): False
                      )
                    (eval): DataLoaderCreator(
                        (seed): 0
                        (batch_size): 2
                        (shuffle): False
                      )
                )
              (model): DummyClassificationModel(
                  (linear): Linear(in_features=4, out_features=3, bias=True)
                  (criterion): CrossEntropyLoss()
                )
              (optimizer): SGD (
                Parameter Group 0
                    dampening: 0
                    differentiable: False
                    foreach: None
                    lr: 0.01
                    maximize: False
                    momentum: 0
                    nesterov: False
                    weight_decay: 0
                )
              (lr_scheduler): None
              (training_loop): TrainingLoop(
                  (set_grad_to_none): True
                  (batch_device_placement): AutoDevicePlacement(device=cpu)
                  (tag): train
                  (clip_grad_fn): None
                  (clip_grad_args): ()
                  (observer): NoOpLoopObserver()
                  (profiler): NoOpProfiler()
                )
              (evaluation_loop): EvaluationLoop(
                  (batch_device_placement): AutoDevicePlacement(device=cpu)
                  (grad_enabled): False
                  (tag): eval
                  (condition): EveryEpochEvalCondition(every=1)
                  (observer): NoOpLoopObserver()
                  (profiler): NoOpProfiler()
                )
              (state): EngineState(
                  (modules): AssetManager(num_assets=5)
                  (histories): HistoryManager()
                  (random_seed): 9984043075503325450
                  (max_epochs): 1
                  (epoch): -1
                  (iteration): -1
                )
              (event_manager): EventManager(
                  (event_handlers):
                  (last_triggered_event): None
                )
              (exp_tracker): NoOpExpTracker(experiment_path=..., is_activated=True))
          (exp_tracker): None
          (handlers): ()
          (resources): ()
          (random_seed): 10139531598155730726
        )
        >>> runner.run()
    """

    def __init__(
        self,
        engine: BaseEngine | dict,
        handlers: Sequence[BaseHandler | dict] | None = None,
        exp_tracker: BaseExpTracker | dict | None = None,
        random_seed: int = 10139531598155730726,
        resources: Sequence[BaseResource | dict] | None = None,
    ) -> None:
        super().__init__(resources=resources)
        self._engine = engine
        self._handlers = () if handlers is None else tuple(handlers)
        self._exp_tracker = exp_tracker
        self._random_seed = random_seed

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping(
                {
                    "engine": self._engine,
                    "exp_tracker": self._exp_tracker,
                    "handlers": "\n" + str_sequence(self._handlers) if self._handlers else "()",
                    "resources": "\n" + str_sequence(self._resources) if self._resources else "()",
                    "random_seed": self._random_seed,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def _run(self) -> BaseEngine:
        return _run_training_pipeline(
            engine=self._engine,
            handlers=self._handlers,
            exp_tracker=self._exp_tracker,
            random_seed=self._random_seed,
        )


def _run_training_pipeline(
    engine: BaseEngine | dict,
    handlers: tuple[BaseHandler | dict, ...] | list[BaseHandler | dict],
    exp_tracker: BaseExpTracker | dict | None,
    random_seed: int = 4398892194000378040,
) -> BaseEngine:
    r"""Implements the training pipeline.

    Internally, this function does the following steps:

        - set the experiment tracker
        - set the random seed
        - instantiate the engine
        - set up and attach the handlers
        - train the model with the engine

    Args:
    ----
        engine (``BaseEngine`` or dict): Specifies the engine or its
            configuration.
        handlers (list or tuple): Specifies the list of handlers or
            their configuration.
        exp_tracker (``BaseExpTracker`` or dict or None): Specifies
            the experiment tracker or its configuration. If ``None``,
            the no-operation experiment tracker is used.
        random_seed (int, optional): Specifies the random seed.
            Default: ``4398892194000378040``

    Returns:
    -------
        ``BaseEngine``: The trained engine.
    """
    with setup_exp_tracker(exp_tracker) as tracker:
        random_seed = random_seed + dist.get_rank()
        logger.info(f"Set the random seed to {random_seed}")
        manual_seed(random_seed)

        if isinstance(engine, dict):
            tracker.log_hyper_parameters(engine)
            logger.info("Initializing the engine from its configuration...")
            engine = BaseEngine.factory(exp_tracker=tracker, **engine)

        logger.info("Adding the handlers to the engine...")
        setup_and_attach_handlers(engine, handlers)

        logger.info(f"engine:\n{engine}")
        engine.train()

    return engine



from __future__ import annotations

__all__ = ["ModelFreezer"]

import logging
from typing import TYPE_CHECKING

from gravitorch.engines.events import EngineEvents
from gravitorch.handlers.base import BaseHandler
from gravitorch.handlers.utils import add_unique_event_handler
from gravitorch.nn import freeze_module
from gravitorch.utils.events import GEventHandler

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine

logger = logging.getLogger(__name__)


class ModelFreezer(BaseHandler):
    r"""Implements a handler to freeze the model or one of its
    submodules.

    Args:
    ----
        event (str, optional): Specifies the event when the model
            is frozen. Default: ``'train_started'``
        module_name (str, optional): Specifies the name of the module
            to freeze if only one of the submodules should be frozen.
            Default: ``''``

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.handlers import ModelFreezer
        >>> from gravitorch.testing import create_dummy_engine
        >>> engine = create_dummy_engine()
        >>> engine.model
        DummyClassificationModel(
          (linear): Linear(in_features=4, out_features=3, bias=True)
          (criterion): CrossEntropyLoss()
        )
        >>> handler = ModelFreezer()
        >>> handler
        ModelFreezer(event=train_started, module_name=)
        >>> handler.attach(engine)
        >>> engine.trigger_event("train_started")
        >>> for name, param in engine.model.named_parameters():
        ...     print(name, param.requires_grad)
        ...
        linear.weight False
        linear.bias False
    """

    def __init__(
        self,
        event: str = EngineEvents.TRAIN_STARTED,
        module_name: str = "",
    ) -> None:
        self._module_name = str(module_name)
        self._event = str(event)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(event={self._event}, module_name={self._module_name})"
        )

    def attach(self, engine: BaseEngine) -> None:
        add_unique_event_handler(
            engine=engine,
            event=self._event,
            event_handler=GEventHandler(
                self.freeze,
                handler_kwargs={"engine": engine},
            ),
        )

    def freeze(self, engine: BaseEngine) -> None:
        r"""Freezes the model or one of its submodules.

        Args:
        ----
            engine (``BaseEngine``): Specifies the engine with the
                model.

        Example usage:

        .. code-block:: pycon

            >>> from gravitorch.handlers import ModelFreezer
            >>> from gravitorch.testing import create_dummy_engine
            >>> engine = create_dummy_engine()
            >>> engine.model
            DummyClassificationModel(
              (linear): Linear(in_features=4, out_features=3, bias=True)
              (criterion): CrossEntropyLoss()
            )
            >>> handler = ModelFreezer()
            >>> handler.freeze(engine)
            >>> for name, param in engine.model.named_parameters():
            ...     print(name, param.requires_grad)
            ...
            linear.weight False
            linear.bias False
        """
        if self._module_name:
            logger.info(f"Freeze submodule {self._module_name}")
        else:
            logger.info("Freeze model")
        freeze_module(engine.model.get_submodule(self._module_name))



r"""This package contains the model base class and some implemented
models."""

from __future__ import annotations

__all__ = [
    "BaseModel",
    "VanillaModel",
    "is_model_config",
    "setup_and_attach_model",
    "setup_model",
]

from gravitorch.models.base import BaseModel
from gravitorch.models.utils import is_model_config, setup_and_attach_model, setup_model
from gravitorch.models.vanilla import VanillaModel



r"""This module defines some utility functions for the data loader
creators."""

from __future__ import annotations

__all__ = ["is_dataloader_creator_config", "setup_dataloader_creator"]

import logging

from objectory.utils import is_object_config

from gravitorch.creators.dataloader.base import BaseDataLoaderCreator
from gravitorch.creators.dataloader.pytorch import AutoDataLoaderCreator
from gravitorch.utils.format import str_target_object

logger = logging.getLogger(__name__)


def is_dataloader_creator_config(config: dict) -> bool:
    r"""Indicates if the input configuration is a configuration for a
    ``BaseDataLoaderCreator``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
    ----
        config (dict): Specifies the configuration to check.

    Returns:
    -------
        bool: ``True`` if the input configuration is a configuration
            for a ``BaseDataLoaderCreator`` object.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.creators.dataloader import is_dataloader_creator_config
        >>> is_dataloader_creator_config(
        ...     {"_target_": "gravitorch.creators.dataloader.DataLoaderCreator"}
        ... )
        True
    """
    return is_object_config(config, BaseDataLoaderCreator)


def setup_dataloader_creator(
    creator: BaseDataLoaderCreator | dict | None,
) -> BaseDataLoaderCreator:
    r"""Sets up a data loader creator.

    Args:
    ----
        creator (``BaseDataLoaderCreator`` or dict or None):
            Specifies the data loader creator or its configuration.
            If ``None``, a data loader creator will be created
            automatically.

    Returns:
    -------
        ``BaseDataLoaderCreator``: The data loader creator.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.creators.dataloader import setup_dataloader_creator
        >>> from gravitorch.testing import DummyDataset
        >>> creator = setup_dataloader_creator(
        ...     {"_target_": "gravitorch.creators.dataloader.DataLoaderCreator"}
        ... )
        >>> creator
        DataLoaderCreator(
          (seed): 0
        )
    """
    if creator is None:
        creator = AutoDataLoaderCreator()
    if isinstance(creator, dict):
        logger.info(
            "Initializing a data loader creator from its configuration... "
            f"{str_target_object(creator)}"
        )
        creator = BaseDataLoaderCreator.factory(**creator)
    if not isinstance(creator, BaseDataLoaderCreator):
        logger.warning(f"creator is not a `BaseDataLoaderCreator` (received: {type(creator)})")
    return creator



r"""This module defines the base class to implement evaluation loops."""

from __future__ import annotations

__all__ = ["BaseEvaluationLoop"]

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

from objectory import AbstractFactory

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine

logger = logging.getLogger(__name__)


class BaseEvaluationLoop(ABC, metaclass=AbstractFactory):
    r"""Defines the evaluation loop base class.

    To implement your own evaluation loop, you will need to define the
    following methods:

        - ``eval``
        - ``load_state_dict``
        - ``state_dict``

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.loops.evaluation import EvaluationLoop
        >>> from gravitorch.testing import create_dummy_engine
        >>> engine = create_dummy_engine()
        >>> loop = EvaluationLoop()
        >>> loop
        EvaluationLoop(
          (batch_device_placement): AutoDevicePlacement(device=cpu)
          (grad_enabled): False
          (tag): eval
          (condition): EveryEpochEvalCondition(every=1)
          (observer): NoOpLoopObserver()
          (profiler): NoOpProfiler()
        )
        >>> loop.eval(engine)
    """

    @abstractmethod
    def eval(self, engine: BaseEngine) -> None:
        r"""Evaluates the model on the evaluation dataset.

        Args:
        ----
            engine (``BaseEngine``): Specifies the engine.

        Example usage:

        .. code-block:: pycon

            >>> from gravitorch.loops.evaluation import EvaluationLoop
            >>> from gravitorch.testing import create_dummy_engine
            >>> engine = create_dummy_engine()
            >>> loop = EvaluationLoop()
            >>> loop.eval(engine)
        """

    @abstractmethod
    def load_state_dict(self, state_dict: dict[str, Any]) -> None:
        r"""Sets up the evaluation loop from a dictionary containing the
        state values.

        Args:
        ----
            state_dict (dict): Specifies a dictionary
                containing state keys with values.

        Example usage:

        .. code-block:: pycon

            >>> from gravitorch.loops.evaluation import EvaluationLoop
            >>> loop = EvaluationLoop()
            >>> loop.load_state_dict({})
        """

    @abstractmethod
    def state_dict(self) -> dict[str, Any]:
        r"""Returns a dictionary containing state values.

        Returns
        -------
            dict: The state values in a dict.

        Example usage:

        .. code-block:: pycon

            >>> from gravitorch.loops.evaluation import EvaluationLoop
            >>> loop = EvaluationLoop()
            >>> state = loop.state_dict()
            >>> state
            {}
        """



__all__ = ["BaseDataStream", "DataLoaderDataStream", "IterableDataStream"]

from gravitorch.datastreams.base import BaseDataStream
from gravitorch.datastreams.dataloader import DataLoaderDataStream
from gravitorch.datastreams.iterable import IterableDataStream



from __future__ import annotations

__all__ = ["BaseOptimizerCreator"]

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from objectory import AbstractFactory
from torch.nn import Module
from torch.optim import Optimizer

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine


class BaseOptimizerCreator(ABC, metaclass=AbstractFactory):
    r"""Defines the base class to create an optimizer.

    Note that it is not the unique approach to create an optimizer. Feel
    free to use other approaches if this approach does not fit your
    needs.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.testing import create_dummy_engine, DummyClassificationModel
        >>> from gravitorch.creators.optimizer import OptimizerCreator
        >>> creator = OptimizerCreator({"_target_": "torch.optim.SGD", "lr": 0.01})
        >>> creator
        OptimizerCreator(add_module_to_engine=True)
        >>> engine = create_dummy_engine()
        >>> model = DummyClassificationModel()
        >>> optimizer = creator.create(engine, model)
        >>> optimizer
        SGD (
        Parameter Group 0...
            lr: 0.01
            maximize: False
            momentum: 0
            nesterov: False
            weight_decay: 0
        )
    """

    @abstractmethod
    def create(self, engine: BaseEngine, model: Module) -> Optimizer | None:
        r"""Creates an optimizer.

        This method is responsible to register the event handlers
        associated to the optimizer.

        Args:
        ----
            engine (``gravitorch.engines.BaseEngine``): Specifies an
                engine.
            model (``torch.nn.Module``): Specifies a model.

        Returns:
        -------
            ``torch.optim.Optimizer`` or ``None``: The created
                optimizer or ``None`` if no optimizer is created.

        Example usage:

        .. code-block:: pycon

            >>> from gravitorch.testing import create_dummy_engine, DummyClassificationModel
            >>> from gravitorch.creators.optimizer import OptimizerCreator
            >>> creator = OptimizerCreator({"_target_": "torch.optim.SGD", "lr": 0.01})
            >>> engine = create_dummy_engine()
            >>> model = DummyClassificationModel()
            >>> optimizer = creator.create(engine, model)
            >>> optimizer
            SGD (
            Parameter Group 0...
                lr: 0.01
                maximize: False
                momentum: 0
                nesterov: False
                weight_decay: 0
            )
        """



from __future__ import annotations

__all__ = ["BaseDatasetCreator", "DatasetCreator", "setup_dataset_creator"]

from gravitorch.creators.dataset.base import BaseDatasetCreator, setup_dataset_creator
from gravitorch.creators.dataset.vanilla import DatasetCreator



from __future__ import annotations

__all__ = ["DictBatcherIterDataPipeCreator"]

from collections.abc import Sequence
from typing import TYPE_CHECKING, TypeVar

from torch import Tensor
from torch.utils.data import IterDataPipe, MapDataPipe

from gravitorch.creators.datapipe.base import BaseDataPipeCreator
from gravitorch.data.datacreators.base import BaseDataCreator, setup_datacreator
from gravitorch.datapipes.iter import DictBatcher
from gravitorch.utils.format import str_indent, str_mapping

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine


T = TypeVar("T")


class DictBatcherIterDataPipeCreator(BaseDataPipeCreator[T]):
    r"""Create a ``DictBatcher`` where the data is generated by a
    ``BaseDataCreator`` object.

    Args:
    ----
        data (``BaseDataCreator`` or ``dict``): Specifies the data
            creator or its configuration.
        **kwargs: See documentation of ``DictBatcher``

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from gravitorch.data.datacreators import DataCreator
        >>> from gravitorch.creators.datapipe import DictBatcherIterDataPipeCreator
        >>> creator = DictBatcherIterDataPipeCreator(
        ...     DataCreator({"key": torch.ones(6, 3)}), batch_size=4
        ... )
        >>> datapipe = creator.create()
        >>> datapipe
        DictBatcherIterDataPipe(
        (batch_size): 4
          (shuffle): False
          (random_seed): 11918852809641073385
          (datapipe_or_data): <class 'dict'> (length=1)
              (key): <class 'torch.Tensor'> | shape=torch.Size([6, 3]) | dtype=torch.float32 | device=cpu
        )
    """

    def __init__(self, data: BaseDataCreator[dict[str, Tensor]] | dict, **kwargs) -> None:
        self._data = setup_datacreator(data)
        self._kwargs = kwargs

    def __repr__(self) -> str:
        config = {"data": self._data} | self._kwargs
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  {str_indent(str_mapping(config, sorted_keys=True))}\n)"
        )

    def create(
        self, engine: BaseEngine | None = None, source_inputs: Sequence | None = None
    ) -> IterDataPipe[T] | MapDataPipe[T]:
        return DictBatcher(self._data.create(engine), **self._kwargs)



from __future__ import annotations

__all__ = ["DataLoader2Creator", "VanillaDataLoader2Creator"]

import logging
from collections.abc import Iterable
from typing import TYPE_CHECKING, TypeVar

from coola.utils import str_indent, str_mapping
from torch.utils.data import IterDataPipe, MapDataPipe

from gravitorch.creators.dataloader2.base import BaseDataLoader2Creator
from gravitorch.creators.datapipe.base import (
    BaseDataPipeCreator,
    setup_datapipe_creator,
)
from gravitorch.creators.datapipe.vanilla import DataPipeCreator
from gravitorch.dataloaders import create_dataloader2, setup_dataloader2
from gravitorch.datapipes.factory import is_datapipe_config
from gravitorch.utils.imports import is_torchdata_available

if is_torchdata_available():
    from torchdata.dataloader2 import DataLoader2, ReadingServiceInterface
    from torchdata.dataloader2.adapter import Adapter
else:  # pragma: no cover
    Adapter = "Adapter"
    DataLoader2 = "DataLoader2"
    ReadingServiceInterface = "ReadingServiceInterface"

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine

T = TypeVar("T")

logger = logging.getLogger(__name__)


class DataLoader2Creator(BaseDataLoader2Creator[T]):
    r"""Implements a simple dataloader creator.

    Args:
    ----
        dataloader (``torchdata.dataloader2.DataLoader2`` or dict):
            Specifies the dataloader or its configuration.
        cache (bool, optional): If ``True``, the dataloader is created
            only the first time, and then the same data is returned
            for each call to the ``create`` method.
            Default: ``False``

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.creators.dataloader2 import DataLoader2Creator
        >>> from torch.utils.data.datapipes.iter import IterableWrapper
        >>> from torchdata.dataloader2 import DataLoader2
        >>> creator = DataLoader2Creator(
        ...     {
        ...         "_target_": "torchdata.dataloader2.DataLoader2",
        ...         "datapipe": IterableWrapper((1, 2, 3, 4)),
        ...     },
        ... )
        >>> creator.create()
        <torchdata.dataloader2.dataloader2.DataLoader2 object at 0x...>
    """

    def __init__(self, dataloader: DataLoader2 | dict, cache: bool = False) -> None:
        self._dataloader = dataloader
        self._cache = bool(cache)

    def __repr__(self) -> str:
        config = {"dataloader": self._dataloader, "cache": self._cache}
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  {str_indent(str_mapping(config, sorted_keys=True))}\n)"
        )

    def create(self, engine: BaseEngine | None = None) -> DataLoader2[T]:
        dataloader = setup_dataloader2(self._dataloader)
        if self._cache:
            self._dataloader = dataloader
        return dataloader


class VanillaDataLoader2Creator(BaseDataLoader2Creator[T]):
    r"""Implements a simple dataloader creator.

    Args:
    ----
        datapipe (``IterDataPipe`` or ``MapDataPipe`` or
            ``BaseDataPipeCreator`` or dict): Specifies a
            datapipe (or its configuration) or a datapipe creator
            (or its configuration).
        datapipe_adapter_fn: Specifies the ``Adapter`` function(s)
            that will be applied to the DataPipe. Default: ``None``
        reading_service: Defines how ``DataLoader2`` should execute
            operations over the ``DataPipe``, e.g.
            multiprocessing/distributed. Default: ``None``

    Example usage:

    .. code-block:: pycon

        >>> from torch.utils.data.datapipes.iter import IterableWrapper
        >>> from gravitorch.creators.dataloader2 import VanillaDataLoader2Creator
        >>> creator = VanillaDataLoader2Creator(IterableWrapper([1, 2, 3, 4, 5]))
        >>> creator.create()
        <torchdata.dataloader2.dataloader2.DataLoader2 object at 0x...>
    """

    def __init__(
        self,
        datapipe: IterDataPipe[T] | MapDataPipe[T] | BaseDataPipeCreator[T] | dict,
        datapipe_adapter_fn: Iterable[Adapter | dict] | Adapter | dict | None = None,
        reading_service: ReadingServiceInterface | dict | None = None,
    ) -> None:
        if isinstance(datapipe, (IterDataPipe, MapDataPipe)) or (
            isinstance(datapipe, dict) and is_datapipe_config(datapipe)
        ):
            datapipe = DataPipeCreator(datapipe)
        self._datapipe = setup_datapipe_creator(datapipe)
        self._datapipe_adapter_fn = datapipe_adapter_fn
        self._reading_service = reading_service

    def __repr__(self) -> str:
        config = {
            "datapipe": self._datapipe,
            "datapipe_adapter_fn": self._datapipe_adapter_fn,
            "reading_service": self._reading_service,
        }
        return f"{self.__class__.__qualname__}(\n  {str_indent(str_mapping(config))}\n)"

    def create(self, engine: BaseEngine | None = None) -> DataLoader2[T]:
        datapipe = self._datapipe.create(engine)
        logger.info(f"datapipe:\n{datapipe}")
        return create_dataloader2(
            datapipe=datapipe,
            datapipe_adapter_fn=self._datapipe_adapter_fn,
            reading_service=self._reading_service,
        )



from __future__ import annotations

__all__ = ["NoOpInitializer"]

import logging

from torch.nn import Module

from gravitorch.nn.init.base import BaseInitializer

logger = logging.getLogger(__name__)


class NoOpInitializer(BaseInitializer):
    r"""This is the special class that does not update the module
    parameters.

    You should use this class if the parameters of the module are
    initialized somewhere else.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.nn.init import NoOpInitializer
        >>> from torch import nn
        >>> module = nn.Sequential(nn.Linear(4, 6), nn.ReLU(), nn.BatchNorm1d(6), nn.Linear(6, 1))
        >>> initializer = NoOpInitializer()
        >>> initializer
        NoOpInitializer()
        >>> initializer.initialize(module)
    """

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def initialize(self, module: Module) -> None:
        logger.info("The module parameters are not updated")



r"""This module defines the base class for the evaluation loops."""

from __future__ import annotations

__all__ = ["BaseBasicEvaluationLoop"]

import logging
import sys
from abc import abstractmethod
from collections.abc import Iterable
from typing import Any

from torch.nn import Module

from gravitorch import constants as ct
from gravitorch.datastreams.base import BaseDataStream
from gravitorch.distributed import comm as dist
from gravitorch.engines.base import BaseEngine
from gravitorch.engines.events import EngineEvents
from gravitorch.loops.evaluation.base import BaseEvaluationLoop
from gravitorch.loops.evaluation.conditions import (
    BaseEvalCondition,
    EveryEpochEvalCondition,
)
from gravitorch.loops.observers.base import BaseLoopObserver
from gravitorch.loops.observers.factory import setup_loop_observer
from gravitorch.utils.history import MinScalarHistory
from gravitorch.utils.imports import is_tqdm_available
from gravitorch.utils.metric_tracker import ScalarMetricTracker
from gravitorch.utils.profilers import BaseProfiler, setup_profiler
from gravitorch.utils.seed import manual_seed
from gravitorch.utils.timing import BatchIterTimer

if is_tqdm_available():
    from tqdm import tqdm
else:  # pragma: no cover

    def tqdm(x: Iterable, *args, **kwargs) -> Iterable:
        return x


logger = logging.getLogger(__name__)


class BaseBasicEvaluationLoop(BaseEvaluationLoop):
    r"""Implements a simple evaluation loop to evaluate a model on a
    given dataset.

    Args:
    ----
        tag (str, optional): Specifies the tag which is used to log
            metrics. Default: ``"eval"``
        condition (``BaseEvalCondition`` or dict or None): Specifies
            the condition to evaluate the loop or its configuration.
            If ``None``, the ``EveryEpochEvalCondition(every=1)`` is
            used.  Default ``None``
        observer (``BaseLoopObserver`` or dict or None, optional):
            Specifies the loop observer or its configuration.
            If ``None``, the ``NoOpLoopObserver`` is instantiated.
            Default: ``None``
        profiler (``BaseProfiler`` or dict or None, optional):
            Specifies the profiler or its configuration. If ``None``,
            the ``NoOpProfiler`` is instantiated. Default: ``None``
    """

    def __init__(
        self,
        tag: str = "eval",
        condition: BaseEvalCondition | dict | None = None,
        observer: BaseLoopObserver | dict | None = None,
        profiler: BaseProfiler | dict | None = None,
    ) -> None:
        self._tag = str(tag)
        self._condition = self._setup_condition(condition)
        logger.info(f"condition:\n{self._condition}")
        self._observer = setup_loop_observer(observer)
        logger.info(f"observer:\n{self._observer}")
        self._profiler = setup_profiler(profiler)
        logger.info(f"profiler:\n{self._profiler}")

    def eval(self, engine: BaseEngine) -> None:
        dist.barrier()
        if not engine.datasource.has_iterable(self._tag) or not self._condition(engine):
            return
        logger.info(f"Evaluating model for epoch {engine.epoch}")

        self._prepare_evaluation(engine)
        dist.barrier()
        engine.trigger_event(EngineEvents.EVAL_EPOCH_STARTED)

        model, datastream = self._prepare_model_datastream(engine)
        dist.barrier()

        self._observer.start(engine)
        with datastream as dataiter:
            self._evaluate_loop(engine=engine, model=model, dataiter=dataiter)
        self._observer.end(engine)

        engine.trigger_event(EngineEvents.EVAL_EPOCH_COMPLETED)
        dist.barrier()

    def load_state_dict(self, state_dict: dict[str, Any]) -> None:
        pass

    def state_dict(self) -> dict[str, Any]:
        return {}

    def _prepare_evaluation(self, engine: BaseEngine) -> None:
        r"""Prepares the evaluation.

        Args:
        ----
            engine (``BaseEngine``): Specifies the engine.
        """
        # Fix the random seed for reproducibility purpose.
        manual_seed(engine.random_seed + engine.epoch + engine.max_epochs * dist.get_rank())
        engine.model.eval()

        if not engine.has_history(f"{self._tag}/{ct.LOSS}"):
            engine.add_history(MinScalarHistory(f"{self._tag}/{ct.LOSS}"))

    def _evaluate_loop(self, engine: BaseEngine, model: Module, dataiter: Iterable) -> None:
        r"""Computes the evaluation loop.

        Args:
        ----
            engine (``BaseEngine``): Specifies the engine.
            model (``torch.nn.Module``): Specifies the model.
            dataiter (``Iterable``): Specifies the iterable on data.
        """
        metrics = ScalarMetricTracker()
        prefix = f"({dist.get_rank()}/{dist.get_world_size()}) " if dist.is_distributed() else ""
        dataiter = tqdm(
            dataiter,
            desc=f"{prefix}Evaluation [{engine.epoch}]",
            position=dist.get_rank(),
            file=sys.stdout,
        )
        dataiter = BatchIterTimer(dataiter, epoch=engine.epoch, prefix=f"{self._tag}/")
        dist.barrier()

        with self._profiler as profiler:
            for batch in dataiter:
                output = self._eval_one_batch(engine, model, batch)
                metrics.update(output)
                self._observer.update(engine=engine, model_input=batch, model_output=output)
                profiler.step()

        # To be sure the progress bar is displayed before the following lines
        sys.stdout.flush()
        dist.barrier()

        # Log some evaluation metrics to the engine.
        dataiter.log_stats(engine=engine)
        metrics.log_average_value(engine=engine, prefix=f"{self._tag}/")
        dist.barrier()

    def _setup_condition(self, condition: BaseEvalCondition | dict | None) -> BaseEvalCondition:
        r"""Sets up the condition.

        The condition is instantiated from its configuration by using
        the ``BaseEvalCondition`` factory function.

        Args:
        ----
            condition (``BaseEvalCondition`` or dict or None):
                Specifies the condition or its configuration.
                If ``None``, the ``EveryEpochEvalCondition(every=1)``
                is instantiated.

        Returns:
        -------
            ``BaseEvalCondition``: The state.
        """
        condition = condition or EveryEpochEvalCondition(every=1)
        if isinstance(condition, dict):
            logger.info("Initializing the condition from its configuration...")
            condition = BaseEvalCondition.factory(**condition)

        logger.info(f"condition:\n{condition}")
        return condition

    @abstractmethod
    def _eval_one_batch(self, engine: BaseEngine, model: Module, batch: Any) -> dict:
        r"""Evaluates the model on the given batch.

        Args:
        ----
            engine (``BaseEngine``): Specifies the engine.
            model (``torch.nn.Module``): Specifies the model to evaluate.
            batch: Specifies the batch of data.

        Returns:
        -------
            dict: Some results about the batch.
        """

    @abstractmethod
    def _prepare_model_datastream(
        self, engine: BaseEngine
    ) -> tuple[Module, BaseDataStream | Iterable]:
        r"""Prepares the model and datastream.

        Args:
        ----
            engine (``BaseEngine``): Specifies the engine.

        Returns:
        -------
            ``torch.nn.Module``, ``BaseDataStream`` or ``Iterable``:
                A tuple with the model and the datastream or data
                iterable.
        """



r"""This module implements some utility functions for the engine
states."""

from __future__ import annotations

__all__ = ["setup_engine_state"]

import logging

from gravitorch.utils.engine_states.base import BaseEngineState
from gravitorch.utils.engine_states.vanilla import EngineState
from gravitorch.utils.format import str_target_object

logger = logging.getLogger(__name__)


def setup_engine_state(state: BaseEngineState | dict | None) -> BaseEngineState:
    r"""Sets up the engine state.

    The state is instantiated from its configuration by using the
    ``BaseEngineState`` factory function.

    Args:
    ----
        state (``BaseEngineState`` or dict or None): Specifies the
            engine state or its configuration. If ``None``, the
            ``EngineState`` is instantiated.

    Returns:
    -------
        ``BaseEngineState``: The engine state.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.utils.engine_states import setup_engine_state
        >>> state = setup_engine_state({"_target_": "gravitorch.utils.engine_states.EngineState"})
        >>> state
        EngineState(
          (modules): AssetManager(num_assets=0)
          (histories): HistoryManager()
          (random_seed): 9984043075503325450
          (max_epochs): 1
          (epoch): -1
          (iteration): -1
        )
    """
    if state is None:
        state = EngineState()
    if isinstance(state, dict):
        logger.info(
            f"Initializing an engine state from its configuration... {str_target_object(state)}"
        )
        state = BaseEngineState.factory(**state)
    return state



from __future__ import annotations

__all__ = ["LRScheduler", "EpochLRScheduler", "IterationLRScheduler"]

from typing import TYPE_CHECKING

from coola.utils import str_indent, str_mapping

from gravitorch.handlers.base import BaseHandler
from gravitorch.handlers.lr_monitor import EpochLRMonitor, IterationLRMonitor
from gravitorch.handlers.lr_scheduler_updater import (
    EpochLRSchedulerUpdater,
    IterationLRSchedulerUpdater,
)
from gravitorch.handlers.utils import setup_handler

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine


class LRScheduler(BaseHandler):
    r"""Implements a handler to update a learning rate (LR) scheduler and
    monitor the LR value.

    Args:
    ----
        lr_scheduler_updater (``BaseHandler`` or dict): Specifies the
            learning rate scheduler updater or its configuration. The
            LR scheduler updater is responsible to update the LR
            scheduler.
        lr_monitor (``BaseHandler`` or dict): Specifies the learning
            rate monitor or its configuration.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.handlers import (
        ...     LRScheduler,
        ...     EpochLRMonitor,
        ...     EpochLRSchedulerUpdater,
        ... )
        >>> from gravitorch.testing import create_dummy_engine
        >>> engine = create_dummy_engine()
        >>> handler = LRScheduler(
        ...     lr_scheduler_updater=EpochLRSchedulerUpdater(), lr_monitor=EpochLRMonitor()
        ... )
        >>> handler
        LRScheduler(
          (lr_monitor): EpochLRMonitor(freq=1, event=train_epoch_started)
          (lr_scheduler_updater): EpochLRSchedulerUpdater(event=train_epoch_completed)
        )
        >>> handler.attach(engine)
        >>> engine.trigger_event("train_epoch_started")
        >>> engine.trigger_event("train_epoch_completed")
    """

    def __init__(
        self,
        lr_scheduler_updater: BaseHandler | dict,
        lr_monitor: BaseHandler | dict,
    ) -> None:
        self._lr_scheduler_updater = setup_handler(lr_scheduler_updater)
        self._lr_monitor = setup_handler(lr_monitor)

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping(
                {
                    "lr_monitor": self._lr_monitor,
                    "lr_scheduler_updater": self._lr_scheduler_updater,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def attach(self, engine: BaseEngine) -> None:
        r"""Attaches the handler to update a LR scheduler and monitor the
        LR value.

        Args:
        ----
            engine (``BaseEngine``): Specifies the engine.
        """
        self._lr_scheduler_updater.attach(engine)
        self._lr_monitor.attach(engine)


class EpochLRScheduler(LRScheduler):
    r"""Implements a handler to update a learning rate (LR) scheduler at
    the end of each training epoch and monitor the LR value.

    This LR scheduler handler sets up:

        - an event handler to update the LR scheduler at the end of
            each training epoch
        - a LR monitor to log the learning rate value(s) at the
            beginning of each training epoch

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.handlers import EpochLRScheduler
        >>> from gravitorch.testing import create_dummy_engine
        >>> engine = create_dummy_engine()
        >>> handler = EpochLRScheduler()
        >>> handler
        EpochLRScheduler(
          (lr_monitor): EpochLRMonitor(freq=1, event=train_epoch_started)
          (lr_scheduler_updater): EpochLRSchedulerUpdater(event=train_epoch_completed)
        )
        >>> handler.attach(engine)
        >>> engine.trigger_event("train_epoch_started")
        >>> engine.trigger_event("train_epoch_completed")
    """

    def __init__(self) -> None:
        super().__init__(
            lr_scheduler_updater=EpochLRSchedulerUpdater(), lr_monitor=EpochLRMonitor()
        )


class IterationLRScheduler(LRScheduler):
    r"""Implements a handler to update a learning rate (LR) scheduler at
    the end of each training iteration and monitor the LR value.

    This LR scheduler handler sets up:

        - an event handler to update the LR scheduler at the end of
            each training iteration
        - a LR monitor to log the learning rate value(s) at the
            beginning of each training iteration

    Args:
    ----
        freq (int, optional): Specifies the iteration frequency used
            to monitor the learning rate. Default: ``10``

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.handlers import IterationLRScheduler
        >>> from gravitorch.testing import create_dummy_engine
        >>> engine = create_dummy_engine()
        >>> handler = IterationLRScheduler()
        >>> handler
        IterationLRScheduler(
          (lr_monitor): IterationLRMonitor(freq=10, event=train_iteration_started)
          (lr_scheduler_updater): IterationLRSchedulerUpdater(event=train_iteration_completed)
        )
        >>> handler.attach(engine)
        >>> engine.trigger_event("train_iteration_started")
        >>> engine.trigger_event("train_iteration_completed")
    """

    def __init__(self, freq: int = 10) -> None:
        super().__init__(
            lr_scheduler_updater=IterationLRSchedulerUpdater(),
            lr_monitor=IterationLRMonitor(freq=freq),
        )



r"""This module implements a handler to initialize model's
parameters."""

from __future__ import annotations

__all__ = ["ModelInitializer"]

import logging
from typing import TYPE_CHECKING

from coola.utils import str_indent, str_mapping

from gravitorch.engines.events import EngineEvents
from gravitorch.handlers.base import BaseHandler
from gravitorch.handlers.utils import add_unique_event_handler
from gravitorch.nn.init import BaseInitializer, setup_initializer
from gravitorch.utils.events import GEventHandler

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine

logger = logging.getLogger(__name__)


class ModelInitializer(BaseHandler):
    r"""Implements a handler to initialize the model's parameters.

    This handler uses a ``BaseInitializer`` object to
    initialize model's parameters.

    Args:
    ----
        initializer (``BaseInitializer`` or dict): Specifies the
            model's parameters initializer or its configuration.
        event (str, optional): Specifies the event when to initialize
            the model's parameters. Default: ``'train_started'``

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.handlers import ModelInitializer
        >>> from gravitorch.nn.init import Constant
        >>> from gravitorch.testing import create_dummy_engine
        >>> engine = create_dummy_engine()
        >>> handler = ModelInitializer(Constant())
        >>> handler
        ModelInitializer(
          (initializer): Constant(value=0.0, learnable_only=True, log_info=False)
          (event): train_started
        )
        >>> handler.attach(engine)
        >>> engine.trigger_event("train_started")
        >>> engine.model.state_dict()
        OrderedDict([('linear.weight', tensor([[0., 0., 0., 0.],
                [0., 0., 0., 0.],
                [0., 0., 0., 0.]])), ('linear.bias', tensor([0., 0., 0.]))])
    """

    def __init__(
        self,
        initializer: BaseInitializer | dict,
        event: str = EngineEvents.TRAIN_STARTED,
    ) -> None:
        self._initializer = setup_initializer(initializer)
        self._event = event

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping(
                {
                    "initializer": self._initializer,
                    "event": self._event,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def attach(self, engine: BaseEngine) -> None:
        add_unique_event_handler(
            engine=engine,
            event=self._event,
            event_handler=GEventHandler(
                self._initializer.initialize,
                handler_kwargs={"module": engine.model},
            ),
        )



from __future__ import annotations

__all__ = ["DistributedDataLoaderCreator"]

from typing import TYPE_CHECKING, TypeVar

from torch.utils.data import DataLoader, Dataset, DistributedSampler

from gravitorch.creators.dataset.base import BaseDatasetCreator, setup_dataset_creator
from gravitorch.creators.dataset.vanilla import DatasetCreator
from gravitorch.dataloaders import create_dataloader
from gravitorch.datasets import is_dataset_config
from gravitorch.distributed import comm as dist
from gravitorch.experimental.dataloader.base import BaseDataLoaderCreator
from gravitorch.utils.format import str_indent, str_mapping
from gravitorch.utils.seed import get_torch_generator

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine

T = TypeVar("T")


class DistributedDataLoaderCreator(BaseDataLoaderCreator[T]):
    r"""Defines a simple distributed PyTorch dataloader creator.

    This dataloader creator uses the ``gravitorch.distributed`` package
    to distribute the example per process. Note that this dataloader
    creator uses the default samplers. If you need a different sampler,
    you will need to implement your own dataloader creator.

    Args:
    ----
        dataset (``torch.utils.data.Dataset``): Specifies a
            dataset (or its configuration) or a dataset creator
            (or its configuration).
        shuffle (bool, optional): Specifies of the examples are
            shuffled or not. You should set to ``True`` to have the
            data reshuffled at every epoch. Default: ``False``
        drop_last (bool, optional): set to ``True`` to drop the last
            incomplete batch, if the dataset size is not divisible by
            the batch size. If ``False`` and the size of dataset is
            not divisible by the batch size, then the last batch will
            be smaller. Default: ``False``
        seed (int, optional): Specifies the random seed used to
            shuffle the samples if ``shuffle=True``. Default: ``0``
        **kwargs: See ``torch.utils.data.DataLoader`` documentation.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.experimental.dataloader import DistributedDataLoaderCreator
        >>> creator = DistributedDataLoaderCreator(
        ...     {
        ...         "_target_": "gravitorch.datasets.DummyMultiClassDataset",
        ...         "num_examples": 10,
        ...         "num_classes": 2,
        ...         "feature_size": 4,
        ...     }
        ... )
        >>> creator.create()
        <torch.utils.data.dataloader.DataLoader object at 0x...>
    """

    def __init__(
        self,
        dataset: Dataset | BaseDatasetCreator | dict,
        shuffle: bool = True,
        drop_last: bool = False,
        seed: int = 0,
        **kwargs,
    ) -> None:
        if isinstance(dataset, Dataset) or (
            isinstance(dataset, dict) and is_dataset_config(dataset)
        ):
            dataset = DatasetCreator(dataset)
        self._dataset = setup_dataset_creator(dataset)
        self._shuffle = bool(shuffle)
        self._drop_last = bool(drop_last)
        self._seed = int(seed)
        self._kwargs = kwargs

    def __repr__(self) -> str:
        config = {
            "dataset": self._dataset,
            "shuffle": self._shuffle,
            "drop_last": self._drop_last,
            "seed": self._seed,
        } | self._kwargs
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  {str_indent(str_mapping(config, sorted_keys=True))}\n)"
        )

    def create(self, engine: BaseEngine | None = None) -> DataLoader[T]:
        dataset = self._dataset.create(engine)
        sampler = DistributedSampler(
            dataset,
            shuffle=self._shuffle,
            drop_last=self._drop_last,
            seed=self._seed,
            rank=dist.get_rank(),
            num_replicas=dist.get_world_size(),
        )
        epoch = 0
        if engine is not None:
            epoch = engine.epoch
            # In distributed mode, calling the set_epoch() method at the beginning
            # of each epoch before creating the DataLoader iterator is necessary to
            # make shuffling work properly across multiple epochs.
            # Otherwise, the same ordering will always be used.
            sampler.set_epoch(epoch)

        # Sampler option is mutually exclusive with shuffle or drop last.
        return create_dataloader(
            dataset,
            sampler=sampler,
            generator=get_torch_generator(self._seed + epoch),
            **self._kwargs,
        )



from __future__ import annotations

__all__ = [
    "AutoDataLoaderCreator",
    "BaseDataLoaderCreator",
    "DataLoaderCreator",
    "DistributedDataLoaderCreator",
    "VanillaDataLoaderCreator",
    "is_dataloader_creator_config",
    "setup_dataloader_creator",
]

from gravitorch.experimental.dataloader.auto import AutoDataLoaderCreator
from gravitorch.experimental.dataloader.base import (
    BaseDataLoaderCreator,
    is_dataloader_creator_config,
    setup_dataloader_creator,
)
from gravitorch.experimental.dataloader.distributed import DistributedDataLoaderCreator
from gravitorch.experimental.dataloader.vanilla import (
    DataLoaderCreator,
    VanillaDataLoaderCreator,
)



from __future__ import annotations

__all__ = ["BaseModelCreator", "is_model_creator_config", "setup_model_creator"]

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from objectory import AbstractFactory
from objectory.utils import is_object_config
from torch import nn

from gravitorch.utils.format import str_target_object

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine

logger = logging.getLogger(__name__)


class BaseModelCreator(ABC, metaclass=AbstractFactory):
    r"""Defines the base class to create a model.

    Note that it is not the unique approach to create a model. Feel
    free to use other approaches if this approach does not fit your
    needs.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.testing import create_dummy_engine
        >>> from gravitorch.creators.model import ModelCreator
        >>> creator = ModelCreator({"_target_": "gravitorch.testing.DummyClassificationModel"})
        >>> creator
        ModelCreator(
          (model_config): {'_target_': 'gravitorch.testing.DummyClassificationModel'}
          (attach_model_to_engine): True
          (add_module_to_engine): True
          (device_placement): AutoDevicePlacement(device=cpu)
        )
        >>> engine = create_dummy_engine()
        >>> model = creator.create(engine)
        >>> model
        DummyClassificationModel(
          (linear): Linear(in_features=4, out_features=3, bias=True)
          (criterion): CrossEntropyLoss()
        )
    """

    @abstractmethod
    def create(self, engine: BaseEngine) -> nn.Module:
        r"""Creates a model on the device(s) where it should run.

        This method is responsible to register the event handlers
        associated to the model. This method is also responsible to
        move the model parameters to the device(s).

        Args:
        ----
            engine (``gravitorch.engines.BaseEngine``): Specifies an
                engine.

        Returns:
        -------
            ``torch.nn.Module``: The created model.

        Example usage:

        .. code-block:: pycon

            >>> from gravitorch.testing import create_dummy_engine
            >>> from gravitorch.creators.model import ModelCreator
            >>> creator = ModelCreator({"_target_": "gravitorch.testing.DummyClassificationModel"})
            >>> engine = create_dummy_engine()
            >>> model = creator.create(engine)
            >>> model
            DummyClassificationModel(
              (linear): Linear(in_features=4, out_features=3, bias=True)
              (criterion): CrossEntropyLoss()
            )
        """


def is_model_creator_config(config: dict) -> bool:
    r"""Indicates if the input configuration is a configuration for a
    ``BaseModelCreator``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
    ----
        config (dict): Specifies the configuration to check.

    Returns:
    -------
        bool: ``True`` if the input configuration is a configuration
            for a ``BaseModelCreator`` object.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.creators.model import is_model_creator_config
        >>> is_model_creator_config(
        ...     {
        ...         "_target_": "gravitorch.creators.model.ModelCreator",
        ...         "model_config": {"_target_": "gravitorch.testing.DummyClassificationModel"},
        ...     }
        ... )
        True
    """
    return is_object_config(config, BaseModelCreator)


def setup_model_creator(creator: BaseModelCreator | dict) -> BaseModelCreator:
    r"""Sets up the model creator.

    The model creator is instantiated from its configuration by using
    the ``BaseModelCreator`` factory function.

    Args:
    ----
        creator (``BaseModelCreator`` or dict): Specifies the model
            creator or its configuration.

    Returns:
    -------
        ``BaseModelCreator``: The instantiated model creator.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.creators.model import setup_model_creator
        >>> creator = setup_model_creator(
        ...     {
        ...         "_target_": "gravitorch.creators.model.ModelCreator",
        ...         "model_config": {"_target_": "gravitorch.testing.DummyClassificationModel"},
        ...     }
        ... )
        >>> creator
        ModelCreator(
          (model_config): {'_target_': 'gravitorch.testing.DummyClassificationModel'}
          (attach_model_to_engine): True
          (add_module_to_engine): True
          (device_placement): AutoDevicePlacement(device=cpu)
        )
    """
    if isinstance(creator, dict):
        logger.info(
            f"Initializing a model creator from its configuration... {str_target_object(creator)}"
        )
        creator = BaseModelCreator.factory(**creator)
    if not isinstance(creator, BaseModelCreator):
        logger.warning(f"creator is not a `BaseModelCreator` (received: {type(creator)})")
    return creator



from __future__ import annotations

__all__ = ["GEventHandler", "GConditionalEventHandler"]

from coola.utils import str_indent, str_mapping
from minevent import ConditionalEventHandler, EventHandler


class GEventHandler(EventHandler):
    r"""Implements a variant of ``minvent.EventHandler`` to not show the
    arguments in the to string method.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.utils.events import GEventHandler
        >>> def hello_handler() -> None:
        ...     print("Hello!")
        ...
        >>> handler = GEventHandler(hello_handler)
        >>> print(repr(handler))  # doctest:+ELLIPSIS
        GEventHandler(
          (handler): <function hello_handler at 0x...>
          (handler_args): ()
          (handler_kwargs): {}
        )
        >>> print(str(handler))  # doctest:+ELLIPSIS
        GEventHandler(
          (handler): <function hello_handler at 0x...>
        )
        >>> handler.handle()
        Hello!
    """

    def __str__(self) -> str:
        args = str_indent(str_mapping({"handler": self._handler}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"


class GConditionalEventHandler(ConditionalEventHandler):
    r"""Implements a variant of ``minvent.GConditionalEventHandler`` to
    not show the arguments in the to string method.

    Example usage:

    .. code-block:: pycon

        >>> from minevent import PeriodicCondition
        >>> from gravitorch.utils.events import GConditionalEventHandler
        >>> def hello_handler() -> None:
        ...     print("Hello!")
        ...
        >>> handler = GConditionalEventHandler(hello_handler, PeriodicCondition(freq=3))
        >>> print(repr(handler))  # doctest:+ELLIPSIS
        GConditionalEventHandler(
          (handler): <function hello_handler at 0x...>
          (handler_args): ()
          (handler_kwargs): {}
          (condition): PeriodicCondition(freq=3, step=0)
        )
        >>> print(str(handler))  # doctest:+ELLIPSIS
        GConditionalEventHandler(
          (handler): <function hello_handler at 0x...>
          (condition): PeriodicCondition(freq=3, step=0)
        )
        >>> handler.handle()
        Hello!
        >>> handler.handle()
        >>> handler.handle()
        >>> handler.handle()
        Hello!
    """

    def __str__(self) -> str:
        args = str_indent(str_mapping({"handler": self._handler, "condition": self._condition}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"



from __future__ import annotations

__all__ = ["BaseDatasetCreator", "setup_dataset_creator"]

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Generic, TypeVar

from objectory import AbstractFactory
from torch.utils.data import Dataset

from gravitorch.utils.format import str_target_object

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine

logger = logging.getLogger(__name__)

T = TypeVar("T")


class BaseDatasetCreator(Generic[T], ABC, metaclass=AbstractFactory):
    r"""Define the base class to implement a dataset creator.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.creators.dataset import DatasetCreator
        >>> creator = DatasetCreator(
        ...     {
        ...         "_target_": "gravitorch.datasets.DummyMultiClassDataset",
        ...         "num_examples": 10,
        ...         "num_classes": 2,
        ...         "feature_size": 4,
        ...     }
        ... )
        >>> creator
        DatasetCreator(
          cache=False
          dataset={'_target_': 'gravitorch.datasets.DummyMultiClassDataset', 'num_examples': 10, 'num_classes': 2, 'feature_size': 4}
        )
        >>> creator.create()
        DummyMultiClassDataset(num_examples=10, num_classes=2, feature_size=4, noise_std=0.2, ...)
    """

    @abstractmethod
    def create(self, engine: BaseEngine | None = None) -> Dataset[T]:
        r"""Create a dataset.

        Args:
        ----
            engine (``gravitorch.engines.BaseEngine`` or ``None``,
                optional): Specifies an engine. Default: ``None``

        Returns:
        -------
            ``torch.utils.data.Dataset``: The created dataset.

        Example usage:

        .. code-block:: pycon

            >>> from gravitorch.creators.dataset import DatasetCreator
            >>> creator = DatasetCreator(
            ...     {
            ...         "_target_": "gravitorch.datasets.DummyMultiClassDataset",
            ...         "num_examples": 10,
            ...         "num_classes": 2,
            ...         "feature_size": 4,
            ...     }
            ... )
            >>> creator.create()
            DummyMultiClassDataset(num_examples=10, num_classes=2, feature_size=4, noise_std=0.2, ...)
        """


def setup_dataset_creator(creator: BaseDatasetCreator | dict) -> BaseDatasetCreator:
    r"""Sets up the dataset creator.

    The dataset creator is instantiated from its configuration by
    using the ``BaseDatasetCreator`` factory function.

    Args:
    ----
        creator (``BaseDatasetCreator`` or dict): Specifies the
            dataset creator or its configuration.

    Returns:
    -------
        ``BaseDatasetCreator``: The instantiated dataset creator.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.creators.dataset import setup_dataset_creator
        >>> creator = setup_dataset_creator(
        ...     {
        ...         "_target_": "gravitorch.creators.dataset.DatasetCreator",
        ...         "dataset": {
        ...             "_target_": "gravitorch.datasets.DummyMultiClassDataset",
        ...             "num_examples": 10,
        ...             "num_classes": 2,
        ...             "feature_size": 4,
        ...         },
        ...     }
        ... )
        >>> creator
        DatasetCreator(
          cache=False
          dataset={'_target_': 'gravitorch.datasets.DummyMultiClassDataset', 'num_examples': 10, 'num_classes': 2, 'feature_size': 4}
        )
    """
    if isinstance(creator, dict):
        logger.info(
            "Initializing the dataset creator from its configuration... "
            f"{str_target_object(creator)}"
        )
        creator = BaseDatasetCreator.factory(**creator)
    return creator



from __future__ import annotations

__all__ = ["DataLoaderCreator", "VanillaDataLoaderCreator"]

from typing import TYPE_CHECKING, TypeVar

from torch.utils.data import DataLoader, Dataset

from gravitorch.creators.dataset import (
    BaseDatasetCreator,
    DatasetCreator,
    setup_dataset_creator,
)
from gravitorch.dataloaders.factory import create_dataloader, setup_dataloader
from gravitorch.datasets.factory import is_dataset_config
from gravitorch.experimental.dataloader.base import BaseDataLoaderCreator
from gravitorch.utils.format import str_indent, str_mapping
from gravitorch.utils.seed import get_torch_generator

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine

T = TypeVar("T")


class DataLoaderCreator(BaseDataLoaderCreator[T]):
    r"""Implements a simple dataloader creator.

    Args:
    ----
        dataloader (``torch.utils.data.DataLoader`` or dict): Specifies
            the dataloader or its configuration.
        cache (bool, optional): If ``True``, the dataloader is created
            only the first time, and then the same data is returned
            for each call to the ``create`` method.
            Default: ``False``

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.experimental.dataloader import DataLoaderCreator
        >>> from gravitorch.datasets import ExampleDataset
        >>> creator = DataLoaderCreator(
        ...     {
        ...         "_target_": "torch.utils.data.DataLoader",
        ...         "dataset": ExampleDataset((1, 2, 3, 4)),
        ...     },
        ... )
        >>> creator.create()
        <torch.utils.data.dataloader.DataLoader object at 0x...>
    """

    def __init__(self, dataloader: DataLoader | dict, cache: bool = False) -> None:
        self._dataloader = dataloader
        self._cache = bool(cache)

    def __repr__(self) -> str:
        config = {"dataloader": self._dataloader, "cache": self._cache}
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  {str_indent(str_mapping(config, sorted_keys=True))}"
            "\n)"
        )

    def create(self, engine: BaseEngine | None = None) -> DataLoader[T]:
        dataloader = setup_dataloader(self._dataloader)
        if self._cache:
            self._dataloader = dataloader
        return dataloader


class VanillaDataLoaderCreator(BaseDataLoaderCreator[T]):
    r"""Implements a simple dataloader creator.

    Args:
    ----
        dataset (``torch.utils.data.Dataset`` or ``BaseDatasetCreator``
            or ``dict``): Specifies a dataset (or its configuration)
            or a dataset creator (or its configuration).
        seed (int, optional): Specifies the random seed used to
            reproduce the shuffling of the samples. Default: ``0``
        **kwargs: See ``torch.utils.data.DataLoader`` documentation.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.experimental.dataloader import VanillaDataLoaderCreator
        >>> creator = VanillaDataLoaderCreator(
        ...     {
        ...         "_target_": "gravitorch.datasets.DummyMultiClassDataset",
        ...         "num_examples": 10,
        ...         "num_classes": 2,
        ...         "feature_size": 4,
        ...     }
        ... )
        >>> creator.create()
        <torch.utils.data.dataloader.DataLoader object at 0x...>
    """

    def __init__(
        self, dataset: Dataset | BaseDatasetCreator | dict, seed: int = 0, **kwargs
    ) -> None:
        if isinstance(dataset, Dataset) or (
            isinstance(dataset, dict) and is_dataset_config(dataset)
        ):
            dataset = DatasetCreator(dataset)
        self._dataset = setup_dataset_creator(dataset)
        self._seed = int(seed)
        self._kwargs = kwargs

    def __repr__(self) -> str:
        config = {"dataset": self._dataset, "seed": self._seed} | self._kwargs
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  {str_indent(str_mapping(config, sorted_keys=True))}\n)"
        )

    def create(self, engine: BaseEngine | None = None) -> DataLoader[T]:
        epoch = 0 if engine is None else engine.epoch
        return create_dataloader(
            self._dataset.create(engine),
            generator=get_torch_generator(self._seed + epoch),
            **self._kwargs,
        )



from __future__ import annotations

__all__ = ["AutoDataLoaderCreator"]

from typing import TYPE_CHECKING, TypeVar

from torch.utils.data import DataLoader, Dataset

from gravitorch.creators.dataset.base import BaseDatasetCreator
from gravitorch.distributed import comm as dist
from gravitorch.experimental.dataloader.base import BaseDataLoaderCreator
from gravitorch.experimental.dataloader.distributed import DistributedDataLoaderCreator
from gravitorch.experimental.dataloader.vanilla import VanillaDataLoaderCreator
from gravitorch.utils.format import str_indent

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine

T = TypeVar("T")


class AutoDataLoaderCreator(BaseDataLoaderCreator[T]):
    r"""Defines a PyTorch dataloader creator that automatically chooses
    the dataloader creator based on the context.

    If the distributed package is activated, it uses the
    ``DistributedDataLoaderCreator``, otherwise it uses
    ``DataLoaderCreator``.

    Note the behavior of this class may change based on the new data
    loader creators.

    Args:
    ----
        dataset (``torch.utils.data.Dataset``): Specifies a
            dataset (or its configuration) or a dataset creator
            (or its configuration).
        **kwargs: See ``DataLoaderCreator`` or
            ``DistributedDataLoaderCreator`` documentation.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.experimental.dataloader import AutoDataLoaderCreator
        >>> creator = AutoDataLoaderCreator(
        ...     {
        ...         "_target_": "gravitorch.datasets.DummyMultiClassDataset",
        ...         "num_examples": 10,
        ...         "num_classes": 2,
        ...         "feature_size": 4,
        ...     }
        ... )
        >>> creator.create()
        <torch.utils.data.dataloader.DataLoader object at 0x...>
    """

    def __init__(self, dataset: Dataset | BaseDatasetCreator | dict, **kwargs) -> None:
        if dist.is_distributed():
            self._dataloader = DistributedDataLoaderCreator(dataset, **kwargs)
        else:
            self._dataloader = VanillaDataLoaderCreator(dataset, **kwargs)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(\n  dataloader={str_indent(self._dataloader)}\n)"

    def create(self, engine: BaseEngine | None = None) -> DataLoader[T]:
        return self._dataloader.create(engine=engine)



from __future__ import annotations

__all__ = [
    "EpochCudaEmptyCache",
    "EpochCudaMemoryMonitor",
    "IterationCudaEmptyCache",
    "IterationCudaMemoryMonitor",
]

import logging
from typing import TYPE_CHECKING

import torch

from gravitorch.engines.events import (
    EngineEvents,
    EpochPeriodicCondition,
    IterationPeriodicCondition,
)
from gravitorch.handlers.base import BaseHandler
from gravitorch.handlers.utils import add_unique_event_handler
from gravitorch.utils.cudamem import log_max_cuda_memory_allocated
from gravitorch.utils.events import GConditionalEventHandler
from gravitorch.utils.exp_trackers import EpochStep, IterationStep

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine

logger = logging.getLogger(__name__)


class EpochCudaMemoryMonitor(BaseHandler):
    r"""Implements a handler to monitor the CUDA memory usage every
    ``freq`` epochs.

    Args:
    ----
        event (str, optional): Specifies the epoch-based event when
            the CUDA memory usage should be captured.
            Default: ``'epoch_completed'``
        freq (int, optional): Specifies the epoch frequency used to
            monitor the CUDA memory usage. Default: ``1``

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.handlers import EpochCudaMemoryMonitor
        >>> from gravitorch.testing import create_dummy_engine
        >>> engine = create_dummy_engine()
        >>> handler = EpochCudaMemoryMonitor()
        >>> handler
        EpochCudaMemoryMonitor(freq=1, event=epoch_completed)
        >>> handler.attach(engine)
        >>> engine.trigger_event("epoch_completed")
    """

    def __init__(self, event: str = EngineEvents.EPOCH_COMPLETED, freq: int = 1) -> None:
        self._event = str(event)
        if freq < 1:
            raise ValueError(f"freq has to be greater than 0 (received: {freq:,})")
        self._freq = int(freq)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(freq={self._freq}, event={self._event})"

    def attach(self, engine: BaseEngine) -> None:
        add_unique_event_handler(
            engine=engine,
            event=self._event,
            event_handler=GConditionalEventHandler(
                self.monitor,
                condition=EpochPeriodicCondition(engine=engine, freq=self._freq),
                handler_kwargs={"engine": engine},
            ),
        )

    def monitor(self, engine: BaseEngine) -> None:
        r"""Monitors the CUDA memory usage.

        Args:
        ----
            engine (``BaseEngine``): Specifies the engine.

        Example usage:

        .. code-block:: pycon

            >>> from gravitorch.handlers import EpochCudaMemoryMonitor
            >>> from gravitorch.testing import create_dummy_engine
            >>> engine = create_dummy_engine()
            >>> handler = EpochCudaMemoryMonitor()
            >>> handler.monitor(engine)
        """
        if torch.cuda.is_available():
            torch.cuda.synchronize()
            log_max_cuda_memory_allocated()
            allocated_memory = torch.cuda.max_memory_allocated()
            total_memory = torch.cuda.mem_get_info()[1]
            engine.log_metrics(
                {
                    "epoch/max_cuda_memory_allocated": allocated_memory,
                    "epoch/max_cuda_memory_allocated_pct": float(allocated_memory / total_memory),
                },
                step=EpochStep(engine.epoch),
            )


class IterationCudaMemoryMonitor(BaseHandler):
    r"""Implements a handler to monitor the CUDA memory usage every
    ``freq`` iterations.

    Args:
    ----
        event (str, optional): Specifies the iteration-based event
            when the CUDA memory usage should be capture.
            Default: ``'epoch_completed'``
        freq (int, optional): Specifies the iteration frequency used
            to monitor the CUDA memory usage. Default: ``1``

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.handlers import IterationCudaMemoryMonitor
        >>> from gravitorch.testing import create_dummy_engine
        >>> engine = create_dummy_engine()
        >>> handler = IterationCudaMemoryMonitor()
        >>> handler
        IterationCudaMemoryMonitor(freq=1, event=train_iteration_completed)
        >>> handler.attach(engine)
        >>> engine.trigger_event("train_iteration_completed")
    """

    def __init__(self, event: str = EngineEvents.TRAIN_ITERATION_COMPLETED, freq: int = 1) -> None:
        self._event = str(event)
        if freq < 1:
            raise ValueError(f"freq has to be greater than 0 (received: {freq:,})")
        self._freq = int(freq)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(freq={self._freq}, event={self._event})"

    def attach(self, engine: BaseEngine) -> None:
        add_unique_event_handler(
            engine=engine,
            event=self._event,
            event_handler=GConditionalEventHandler(
                self.monitor,
                condition=IterationPeriodicCondition(engine=engine, freq=self._freq),
                handler_kwargs={"engine": engine},
            ),
        )

    def monitor(self, engine: BaseEngine) -> None:
        r"""Monitors the CUDA memory usage.

        Args:
        ----
            engine (``BaseEngine``): Specifies the engine.

        Example usage:

        .. code-block:: pycon

            >>> from gravitorch.handlers import IterationCudaMemoryMonitor
            >>> from gravitorch.testing import create_dummy_engine
            >>> engine = create_dummy_engine()
            >>> handler = IterationCudaMemoryMonitor()
            >>> handler.monitor(engine)
        """
        if torch.cuda.is_available():
            torch.cuda.synchronize()
            log_max_cuda_memory_allocated()
            allocated_memory = torch.cuda.max_memory_allocated()
            total_memory = torch.cuda.mem_get_info()[1]
            engine.log_metrics(
                {
                    "iteration/max_cuda_memory_allocated": allocated_memory,
                    "iteration/max_cuda_memory_allocated_pct": float(
                        allocated_memory / total_memory
                    ),
                },
                step=IterationStep(engine.iteration),
            )


class EpochCudaEmptyCache(BaseHandler):
    r"""Implements a handler to empty the CUDA cache every ``freq``
    epochs.

    Args:
    ----
        event (str, optional): Specifies the epoch-based event when
            the learning rate should be capture.
            Default: ``'epoch_completed'``
        freq (int, optional): Specifies the epoch frequency used to
            monitor the learning rate. Default: ``1``

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.handlers import EpochCudaEmptyCache
        >>> from gravitorch.testing import create_dummy_engine
        >>> engine = create_dummy_engine()
        >>> handler = EpochCudaEmptyCache()
        >>> handler
        EpochCudaEmptyCache(freq=1, event=epoch_completed)
        >>> handler.attach(engine)
        >>> engine.trigger_event("epoch_completed")
    """

    def __init__(self, event: str = EngineEvents.EPOCH_COMPLETED, freq: int = 1) -> None:
        self._event = str(event)
        if freq < 1:
            raise ValueError(f"freq has to be greater than 0 (received: {freq:,})")
        self._freq = int(freq)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(freq={self._freq}, event={self._event})"

    def attach(self, engine: BaseEngine) -> None:
        add_unique_event_handler(
            engine=engine,
            event=self._event,
            event_handler=GConditionalEventHandler(
                self.empty_cache,
                condition=EpochPeriodicCondition(engine=engine, freq=self._freq),
                handler_kwargs={"engine": engine},
            ),
        )

    def empty_cache(self, engine: BaseEngine) -> None:
        r"""Empty the CUDA cache.

        Args:
        ----
            engine (``BaseEngine``): Specifies the engine.

        Example usage:

        .. code-block:: pycon

            >>> from gravitorch.handlers import EpochCudaEmptyCache
            >>> from gravitorch.testing import create_dummy_engine
            >>> engine = create_dummy_engine()
            >>> handler = EpochCudaEmptyCache()
            >>> handler.empty_cache(engine)
        """
        if torch.cuda.is_available():
            logger.info("Emptying CUDA cache...")
            torch.cuda.empty_cache()


class IterationCudaEmptyCache(BaseHandler):
    r"""Implements a handler to empty the CUDA cache every ``freq``
    iterations.

    Args:
    ----
        event (str, optional): Specifies the iteration-based event when
            the learning rate should be capture.
            Default: ``'train_iteration_completed'``
        freq (int, optional): Specifies the iteration frequency used to
            monitor the learning rate. Default: ``1``

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.handlers import IterationCudaEmptyCache
        >>> from gravitorch.testing import create_dummy_engine
        >>> engine = create_dummy_engine()
        >>> handler = IterationCudaEmptyCache()
        >>> handler
        IterationCudaEmptyCache(freq=1, event=train_iteration_completed)
        >>> handler.attach(engine)
        >>> engine.trigger_event("train_iteration_completed")
    """

    def __init__(self, event: str = EngineEvents.TRAIN_ITERATION_COMPLETED, freq: int = 1) -> None:
        self._event = str(event)
        if freq < 1:
            raise ValueError(f"freq has to be greater than 0 (received: {freq:,})")
        self._freq = int(freq)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(freq={self._freq}, event={self._event})"

    def attach(self, engine: BaseEngine) -> None:
        add_unique_event_handler(
            engine=engine,
            event=self._event,
            event_handler=GConditionalEventHandler(
                self.empty_cache,
                condition=IterationPeriodicCondition(engine=engine, freq=self._freq),
                handler_kwargs={"engine": engine},
            ),
        )

    def empty_cache(self, engine: BaseEngine) -> None:
        r"""Empty the CUDA cache.

        Args:
        ----
            engine (``BaseEngine``): Specifies the engine.

        Example usage:

        .. code-block:: pycon

            >>> from gravitorch.handlers import IterationCudaEmptyCache
            >>> from gravitorch.testing import create_dummy_engine
            >>> engine = create_dummy_engine()
            >>> handler = IterationCudaEmptyCache()
            >>> handler.empty_cache(engine)
        """
        if torch.cuda.is_available():
            logger.info("Emptying CUDA cache...")
            torch.cuda.empty_cache()



from __future__ import annotations

__all__ = ["DataPipeCreator"]

from collections.abc import Sequence
from typing import TYPE_CHECKING, TypeVar

from torch.utils.data import IterDataPipe, MapDataPipe

from gravitorch.creators.datapipe.base import BaseDataPipeCreator
from gravitorch.datapipes.factory import setup_datapipe
from gravitorch.datapipes.utils import clone_datapipe
from gravitorch.utils.format import str_indent, str_mapping

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine

T = TypeVar("T")


class DataPipeCreator(BaseDataPipeCreator[T]):
    r"""Implements a simple ``DataPipe`` creator.

    Args:
    ----
        datapipe (``IterDataPipe`` or ``MapDataPipe`` or dict):
            Specifies the ``DataPipe`` or its configuration.
        cache (bool, optional): If ``True``, the ``DataPipe`` is
            created only the first time, and then the same
            ``DataPipe`` is returned for each call to the
            ``create`` method. Default: ``False``
        deepcopy (bool, optional): If ``True``, the ``DataPipe``
            object is deep-copied before to iterate over the data.
            It allows a deterministic behavior when in-place
            operations are performed on the data.
            Default: ``False``

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.creators.datapipe import DataPipeCreator
        >>> creator = DataPipeCreator(
        ...     {
        ...         "_target_": "torch.utils.data.datapipes.iter.IterableWrapper",
        ...         "iterable": [1, 2, 3, 4],
        ...     }
        ... )
        >>> creator
        DataPipeCreator(
          cache=False
          datapipe={'_target_': 'torch.utils.data.datapipes.iter.IterableWrapper', 'iterable': [1, 2, 3, 4]}
          deepcopy=False
        )
        >>> datapipe = creator.create()
        >>> tuple(datapipe)
        (1, 2, 3, 4)
    """

    def __init__(
        self,
        datapipe: IterDataPipe[T] | MapDataPipe[T] | dict,
        cache: bool = False,
        deepcopy: bool = False,
    ) -> None:
        self._datapipe = datapipe
        self._cache = bool(cache)
        self._deepcopy = bool(deepcopy)

    def __repr__(self) -> str:
        config = {"datapipe": self._datapipe, "cache": self._cache, "deepcopy": self._deepcopy}
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  {str_indent(str_mapping(config, sorted_keys=True))}\n)"
        )

    def create(
        self, engine: BaseEngine | None = None, source_inputs: Sequence | None = None
    ) -> IterDataPipe[T] | MapDataPipe[T]:
        datapipe = setup_datapipe(self._datapipe)
        if self._cache:
            self._datapipe = datapipe
        if self._deepcopy:
            datapipe = clone_datapipe(datapipe, raise_error=False)
        return datapipe



from __future__ import annotations

__all__ = ["BaseDataPipeCreator", "is_datapipe_creator_config", "setup_datapipe_creator"]

import logging
from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import TYPE_CHECKING, Generic, TypeVar

from objectory import AbstractFactory
from objectory.utils import is_object_config
from torch.utils.data import IterDataPipe, MapDataPipe

from gravitorch.utils.format import str_target_object

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine

logger = logging.getLogger(__name__)

T = TypeVar("T")


class BaseDataPipeCreator(Generic[T], ABC, metaclass=AbstractFactory):
    r"""Defines the base class to implement a ``DataPipe`` creator.

    A ``DataPipe`` creator is responsible to create a single
    DataPipe.

    Note: it is possible to create a ``DataPipe`` object without
    using this class.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.creators.datapipe import ChainedDataPipeCreator
        >>> creator = ChainedDataPipeCreator(
        ...     [
        ...         {
        ...             "_target_": "torch.utils.data.datapipes.iter.IterableWrapper",
        ...             "iterable": [1, 2, 3, 4],
        ...         }
        ...     ]
        ... )
        >>> creator
        ChainedDataPipeCreator(
          (0): {'_target_': 'torch.utils.data.datapipes.iter.IterableWrapper', 'iterable': [1, 2, 3, 4]}
        )
        >>> datapipe = creator.create()
        >>> tuple(datapipe)
        (1, 2, 3, 4)
    """

    @abstractmethod
    def create(
        self, engine: BaseEngine | None = None, source_inputs: Sequence | None = None
    ) -> IterDataPipe[T] | MapDataPipe[T]:
        r"""Creates a ``DataPipe`` object.

        Args:
        ----
            engine (``BaseEngine`` or ``None``, optional): Specifies
                an engine. The engine can be used to initialize the
                ``DataPipe`` by using the current epoch value.
                Default: ``None``
            source_inputs (sequence or ``None``): Specifies the first
                positional arguments of the source ``DataPipe``.
                This argument can be used to create a new
                ``DataPipe`` object, that takes existing
                ``DataPipe`` objects as input. See examples below
                to see how to use it. If ``None``, ``source_inputs``
                is set to an empty tuple. Default: ``None``

        Returns:
        -------
            ``IterDataPipe`` or ``MapDataPipe``: The created
                ``DataPipe`` object.

        Example usage:

        .. code-block:: pycon

            >>> from gravitorch.creators.datapipe import ChainedDataPipeCreator
            >>> creator = ChainedDataPipeCreator(
            ...     [
            ...         {
            ...             "_target_": "torch.utils.data.datapipes.iter.IterableWrapper",
            ...             "iterable": [1, 2, 3, 4],
            ...         }
            ...     ]
            ... )
            >>> datapipe = creator.create()
            >>> tuple(datapipe)
            (1, 2, 3, 4)
        """


def is_datapipe_creator_config(config: dict) -> bool:
    r"""Indicates if the input configuration is a configuration for a
    ``BaseDataPipeCreator``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
    ----
        config (dict): Specifies the configuration to check.

    Returns:
    -------
        bool: ``True`` if the input configuration is a configuration
            for a ``BaseDataPipeCreator`` object.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.creators.datapipe import is_datapipe_creator_config
        >>> is_datapipe_creator_config(
        ...     {
        ...         "_target_": "gravitorch.creators.datapipe.ChainedDataPipeCreator",
        ...         "config": [
        ...             {
        ...                 "_target_": "torch.utils.data.datapipes.iter.IterableWrapper",
        ...                 "iterable": [1, 2, 3, 4],
        ...             }
        ...         ],
        ...     }
        ... )
        True
    """
    return is_object_config(config, BaseDataPipeCreator)


def setup_datapipe_creator(creator: BaseDataPipeCreator | dict) -> BaseDataPipeCreator:
    r"""Sets up a ``DataPipe`` creator.

    The ``DataPipe`` creator is instantiated from its
    configuration by using the ``BaseDataPipeCreator`` factory
    function.

    Args:
    ----
        creator (``BaseDataPipeCreator`` or dict): Specifies the
            ``DataPipe`` creator or its configuration.

    Returns:
    -------
        ``BaseDataPipeCreator``: The instantiated ``DataPipe``
            creator.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.creators.datapipe import setup_datapipe_creator
        >>> creator = setup_datapipe_creator(
        ...     {
        ...         "_target_": "gravitorch.creators.datapipe.ChainedDataPipeCreator",
        ...         "config": [
        ...             {
        ...                 "_target_": "torch.utils.data.datapipes.iter.IterableWrapper",
        ...                 "iterable": [1, 2, 3, 4],
        ...             }
        ...         ],
        ...     }
        ... )
        >>> creator
        ChainedDataPipeCreator(
          (0): {'_target_': 'torch.utils.data.datapipes.iter.IterableWrapper', 'iterable': [1, 2, 3, 4]}
        )
    """
    if isinstance(creator, dict):
        logger.info(
            "Initializing a DataPipe creator from its configuration... "
            f"{str_target_object(creator)}"
        )
        creator = BaseDataPipeCreator.factory(**creator)
    if not isinstance(creator, BaseDataPipeCreator):
        logger.warning(f"creator is not a `BaseDataPipeCreator` (received: {type(creator)})")
    return creator



r"""This module contains handlers to update a learning rate (LR)
scheduler."""

from __future__ import annotations

__all__ = [
    "EpochLRSchedulerUpdater",
    "MetricEpochLRSchedulerUpdater",
    "IterationLRSchedulerUpdater",
    "LRSchedulerUpdater",
    "MetricLRSchedulerUpdater",
]

import logging
from typing import TYPE_CHECKING

from gravitorch import constants as ct
from gravitorch.engines.events import EngineEvents
from gravitorch.handlers.base import BaseHandler
from gravitorch.handlers.utils import add_unique_event_handler
from gravitorch.utils.events import GEventHandler

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine

logger = logging.getLogger(__name__)


class LRSchedulerUpdater(BaseHandler):
    r"""Implements a handler to update the learning rate (LR) scheduler
    at a given event.

    Args:
    ----
        event (str): Specifies the event when the learning rate
            scheduler is updated.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.handlers import LRSchedulerUpdater
        >>> from gravitorch.testing import create_dummy_engine
        >>> engine = create_dummy_engine()
        >>> handler = LRSchedulerUpdater("my_event")
        >>> handler
        LRSchedulerUpdater(event=my_event)
        >>> handler.attach(engine)
        >>> engine.trigger_event("my_event")
    """

    def __init__(self, event: str) -> None:
        self._event = str(event)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(event={self._event})"

    def attach(self, engine: BaseEngine) -> None:
        if engine.lr_scheduler is None:
            logger.info("No event is added to the engine because the LR scheduler is not defined")
            return
        add_unique_event_handler(
            engine=engine,
            event=self._event,
            event_handler=GEventHandler(engine.lr_scheduler.step),
        )


class EpochLRSchedulerUpdater(LRSchedulerUpdater):
    r"""Implements a handler to update the learning rate (LR) scheduler
    at the end of each training epoch.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.handlers import EpochLRSchedulerUpdater
        >>> from gravitorch.testing import create_dummy_engine
        >>> engine = create_dummy_engine()
        >>> handler = EpochLRSchedulerUpdater()
        >>> handler
        EpochLRSchedulerUpdater(event=train_epoch_completed)
        >>> handler.attach(engine)
        >>> engine.trigger_event("train_epoch_completed")
    """

    def __init__(self) -> None:
        super().__init__(event=EngineEvents.TRAIN_EPOCH_COMPLETED)


class IterationLRSchedulerUpdater(LRSchedulerUpdater):
    r"""Implements a handler to update the learning rate (LR) scheduler
    at the end of each training iteration.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.handlers import IterationLRSchedulerUpdater
        >>> from gravitorch.testing import create_dummy_engine
        >>> engine = create_dummy_engine()
        >>> handler = IterationLRSchedulerUpdater()
        >>> handler
        IterationLRSchedulerUpdater(event=train_iteration_completed)
        >>> handler.attach(engine)
        >>> engine.trigger_event("train_iteration_completed")
    """

    def __init__(self) -> None:
        super().__init__(event=EngineEvents.TRAIN_ITERATION_COMPLETED)


class MetricLRSchedulerUpdater(BaseHandler):
    r"""Implements a handler to update the learning rate (LR) scheduler
    with a metric value at a given event.

    This updater was designed to be used with ``ReduceLROnPlateau``
    or any LR schedulers that use a metric value as input of the
    ``step`` method.

    Args:
    ----
        event (str): Specifies the event when the learning rate
            scheduler is updated. The event should happen after the
            metric was computed.
        metric_name (str, optional): Specifies the metric to use to
            control the LR scheduler. This metric should be accessible
            with the ``get_history`` method of the engine.
            Default: ``'eval/loss'``

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from gravitorch.handlers import MetricLRSchedulerUpdater
        >>> from gravitorch.testing import DummyClassificationModel, create_dummy_engine
        >>> from gravitorch.utils.exp_trackers import EpochStep
        >>> model = DummyClassificationModel()
        >>> optimizer = torch.optim.SGD(model.parameters(), lr=0.01)
        >>> lr_scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer)
        >>> engine = create_dummy_engine(
        ...     model=model, optimizer=optimizer, lr_scheduler=lr_scheduler
        ... )
        >>> handler = MetricLRSchedulerUpdater("my_event")
        >>> handler
        MetricLRSchedulerUpdater(event=my_event, metric_name=eval/loss)
        >>> handler.attach(engine)
        >>> engine.log_metric("eval/loss", 1.2, step=EpochStep(1))
        >>> engine.trigger_event("my_event")
    """

    def __init__(self, event: str, metric_name: str = f"{ct.EVAL}/loss") -> None:
        self._event = event
        self._metric_name = metric_name

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(event={self._event}, metric_name={self._metric_name})"
        )

    def attach(self, engine: BaseEngine) -> None:
        if engine.lr_scheduler is None:
            logger.info("No event is added to the engine because the LR scheduler is not defined")
            return
        # Note that it is not possible to use the LR scheduler `step` method because the `step`
        # method needs a metric value as input.
        add_unique_event_handler(
            engine=engine,
            event=self._event,
            event_handler=GEventHandler(self.step, handler_kwargs={"engine": engine}),
        )

    def step(self, engine: BaseEngine) -> None:
        r"""Updates the LR scheduler with a metric value.

        Args:
        ----
            engine (``BaseEngine``): Specifies the engine with the LR
                scheduler to update and the metric used to update the
                LR scheduler.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from gravitorch.handlers import MetricLRSchedulerUpdater
            >>> from gravitorch.testing import DummyClassificationModel, create_dummy_engine
            >>> from gravitorch.utils.exp_trackers import EpochStep
            >>> model = DummyClassificationModel()
            >>> optimizer = torch.optim.SGD(model.parameters(), lr=0.01)
            >>> lr_scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer)
            >>> engine = create_dummy_engine(
            ...     model=model, optimizer=optimizer, lr_scheduler=lr_scheduler
            ... )
            >>> handler = MetricLRSchedulerUpdater("my_event")
            >>> engine.log_metric("eval/loss", 1.2, step=EpochStep(1))
            >>> handler.step(engine)
        """
        engine.lr_scheduler.step(engine.get_history(self._metric_name).get_last_value())


class MetricEpochLRSchedulerUpdater(MetricLRSchedulerUpdater):
    r"""Implements a handler to update the learning rate (LR) scheduler
    with a metric value at the end of each epoch.

    This updater was designed to be used with ``ReduceLROnPlateau`` or
    any LR schedulers that use a metric value as input of the ``step``
    method. The ``step`` method of the LR scheduler is called at the
    ``"epoch_completed"`` event, so you need to use a metric that is
    updated before this event.

    Args:
    ----
        metric_name (str, optional): Specifies the metric to use to
            control the LR scheduler. This metric should be accessible
            with the ``get_history`` method of the engine.
            Default: ``'eval/loss'``

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.handlers import MetricEpochLRSchedulerUpdater
        >>> from gravitorch.testing import create_dummy_engine
        >>> engine = create_dummy_engine()
        >>> handler = MetricEpochLRSchedulerUpdater()
        >>> handler
        MetricEpochLRSchedulerUpdater(event=epoch_completed, metric_name=eval/loss)
        >>> handler.attach(engine)
        >>> engine.trigger_event("epoch_completed")
    """

    def __init__(self, metric_name: str = f"{ct.EVAL}/loss") -> None:
        super().__init__(event=EngineEvents.EPOCH_COMPLETED, metric_name=metric_name)



from __future__ import annotations

__all__ = ["DatasetCreator"]

from typing import TYPE_CHECKING, TypeVar

from torch.utils.data import Dataset

from gravitorch.creators.dataset.base import BaseDatasetCreator
from gravitorch.datasets.factory import setup_dataset
from gravitorch.utils.format import str_indent, str_mapping

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine

T = TypeVar("T")


class DatasetCreator(BaseDatasetCreator[T]):
    r"""Implements a simple dataset creator.

    Args:
    ----
        dataset (``torch.utils.data.Dataset`` or dict): Specifies
            the dataset or its configuration.
        cache (bool, optional): If ``True``, the dataset is created
            only the first time, and then the same data is returned
            for each call to the ``create`` method.
            Default: ``False``

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.creators.dataset import DatasetCreator
        >>> creator = DatasetCreator(
        ...     {
        ...         "_target_": "gravitorch.datasets.DummyMultiClassDataset",
        ...         "num_examples": 10,
        ...         "num_classes": 2,
        ...         "feature_size": 4,
        ...     }
        ... )
        >>> creator
        DatasetCreator(
          cache=False
          dataset={'_target_': 'gravitorch.datasets.DummyMultiClassDataset', 'num_examples': 10, 'num_classes': 2, 'feature_size': 4}
        )
        >>> creator.create()
        DummyMultiClassDataset(num_examples=10, num_classes=2, feature_size=4, noise_std=0.2, ...)
    """

    def __init__(self, dataset: Dataset[T] | dict, cache: bool = False) -> None:
        self._dataset = dataset
        self._cache = bool(cache)

    def __repr__(self) -> str:
        config = {"dataset": self._dataset, "cache": self._cache}
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  {str_indent(str_mapping(config, sorted_keys=True))}\n)"
        )

    def create(self, engine: BaseEngine | None = None) -> Dataset[T]:
        dataset = setup_dataset(self._dataset)
        if self._cache:
            self._dataset = dataset
        return dataset



from __future__ import annotations

__all__ = [
    "BaseDataSource",
    "DataCreatorDataSource",
    "DatasetDataSource",
    "DataPipeDataSource",
    "IterableNotFoundError",
    "VanillaDataSource",
    "is_datasource_config",
    "setup_and_attach_datasource",
    "setup_datasource",
]

from gravitorch.datasources.base import (
    BaseDataSource,
    IterableNotFoundError,
    is_datasource_config,
    setup_and_attach_datasource,
    setup_datasource,
)
from gravitorch.datasources.datapipe import DataCreatorDataSource, DataPipeDataSource
from gravitorch.datasources.dataset import DatasetDataSource
from gravitorch.datasources.vanilla import VanillaDataSource



r"""This module implements a simple training loop."""

from __future__ import annotations

__all__ = ["TrainingLoop"]

import logging
from collections.abc import Callable, Iterable
from typing import Any

import torch
from coola.utils import str_indent, str_mapping
from torch.nn import Module
from torch.optim import Optimizer

from gravitorch import constants as ct
from gravitorch.datastreams.base import BaseDataStream
from gravitorch.datastreams.iterable import IterableDataStream
from gravitorch.engines.base import BaseEngine
from gravitorch.engines.events import EngineEvents
from gravitorch.loops.observers import BaseLoopObserver
from gravitorch.loops.training.basic import BaseBasicTrainingLoop
from gravitorch.loops.training.utils import setup_clip_grad
from gravitorch.utils.device_placement import (
    AutoDevicePlacement,
    BaseDevicePlacement,
    setup_device_placement,
)
from gravitorch.utils.profilers import BaseProfiler

logger = logging.getLogger(__name__)


class TrainingLoop(BaseBasicTrainingLoop):
    r"""Implements a simple training loop to train a model on a dataset.

    Args:
    ----
        set_grad_to_none (bool, optional): If ``True``, set the
            gradients to ``None``, otherwise set the gradients to
            zero. Setting the gradients to ``None`` will in general
            have lower memory footprint, and can modestly improve
            performance. Default: ``True``
        batch_device_placement (``BaseDevicePlacement`` or dict or
            ``None``, optional): Specifies the batch device placement
            module. This module moves the batch on a target device.
            The target device should be compatible with the model.
            If ``None``, an ``AutoDevicePlacement`` object is
            instantiated. Default: ``None``
        tag (str, optional): Specifies the tag which is used to log
            metrics. Default: ``"train"``
        clip_grad (dict or None, optional): Specifies the
            configuration to clip the gradient. If ``None``, no
            gradient clipping is used during the training.
            Default: ``None``
        observer (``BaseLoopObserver`` or dict or None, optional):
            Specifies the loop observer or its configuration.
            If ``None``, the ``NoOpLoopObserver`` is instantiated.
            Default: ``None``
        profiler (``BaseProfiler`` or dict or None, optional): Specifies
            the profiler or its configuration. If ``None``, the
            ``NoOpProfiler`` is instantiated. Default: ``None``

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.loops.training import TrainingLoop
        >>> from gravitorch.testing import create_dummy_engine
        >>> engine = create_dummy_engine()
        >>> loop = TrainingLoop()
        >>> loop
        TrainingLoop(
          (set_grad_to_none): True
          (batch_device_placement): AutoDevicePlacement(device=cpu)
          (tag): train
          (clip_grad_fn): None
          (clip_grad_args): ()
          (observer): NoOpLoopObserver()
          (profiler): NoOpProfiler()
        )
        >>> loop.train(engine)
    """

    def __init__(
        self,
        set_grad_to_none: bool = True,
        batch_device_placement: BaseDevicePlacement | dict | None = None,
        tag: str = ct.TRAIN,
        clip_grad: dict | None = None,
        observer: BaseLoopObserver | dict | None = None,
        profiler: BaseProfiler | dict | None = None,
    ) -> None:
        super().__init__(tag=tag, clip_grad=clip_grad, observer=observer, profiler=profiler)
        self._set_grad_to_none = bool(set_grad_to_none)
        self._batch_device_placement = setup_device_placement(
            batch_device_placement or AutoDevicePlacement()
        )

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping(
                {
                    "set_grad_to_none": self._set_grad_to_none,
                    "batch_device_placement": self._batch_device_placement,
                    "tag": self._tag,
                    "clip_grad_fn": self._clip_grad_fn,
                    "clip_grad_args": self._clip_grad_args,
                    "observer": self._observer,
                    "profiler": self._profiler,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def _prepare_model_optimizer_datastream(
        self, engine: BaseEngine
    ) -> tuple[Module, Optimizer, Iterable]:
        logger.info("Preparing the model, optimizer, and datastream...")
        datastream = engine.datasource.get_iterable(iter_id=self._tag, engine=engine)
        datastream = (
            datastream if isinstance(datastream, BaseDataStream) else IterableDataStream(datastream)
        )
        logger.info("Training datastream has been created")
        return engine.model, engine.optimizer, datastream

    def _train_one_batch(
        self, engine: BaseEngine, model: Module, optimizer: Optimizer, batch: Any
    ) -> dict:
        engine.trigger_event(EngineEvents.TRAIN_ITERATION_STARTED)
        optimizer.zero_grad(self._set_grad_to_none)
        output = model(self._batch_device_placement.send(batch))
        engine.trigger_event(EngineEvents.TRAIN_FORWARD_COMPLETED)

        loss = output[ct.LOSS]
        if torch.isnan(loss):
            logger.warning(
                "NaN detected in loss so backpropagation is skipped "
                f"(iteration: {engine.iteration})"
            )
            engine.trigger_event(EngineEvents.TRAIN_ITERATION_COMPLETED)
            return output

        loss.backward()
        if self._clip_grad_fn:
            self._clip_grad_fn(model.parameters(), *self._clip_grad_args)
        engine.trigger_event(EngineEvents.TRAIN_BACKWARD_COMPLETED)

        optimizer.step()
        engine.trigger_event(EngineEvents.TRAIN_ITERATION_COMPLETED)

        return output

    def _setup_clip_grad(self, config: dict) -> tuple[Callable | None, tuple]:
        return setup_clip_grad(config)



from __future__ import annotations

__all__ = ["BaseDataLoader2Creator", "is_dataloader2_creator_config", "setup_dataloader2_creator"]

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Generic, TypeVar

from objectory import AbstractFactory
from objectory.utils import is_object_config

from gravitorch.utils.format import str_target_object
from gravitorch.utils.imports import is_torchdata_available

if is_torchdata_available():
    from torchdata.dataloader2 import DataLoader2
else:  # pragma: no cover
    DataLoader2 = "DataLoader2"

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine


logger = logging.getLogger(__name__)

T = TypeVar("T")


class BaseDataLoader2Creator(Generic[T], ABC, metaclass=AbstractFactory):
    r"""Define the base class to implement a dataloader creator.

    Example usage:

    .. code-block:: pycon

        >>> from torch.utils.data.datapipes.iter import IterableWrapper
        >>> from gravitorch.creators.dataloader2 import VanillaDataLoader2Creator
        >>> creator = VanillaDataLoader2Creator(IterableWrapper([1, 2, 3, 4, 5]))
        >>> creator.create()
        <torchdata.dataloader2.dataloader2.DataLoader2 object at 0x...>
    """

    @abstractmethod
    def create(self, engine: BaseEngine | None = None) -> DataLoader2[T]:
        r"""Create a dataloader.

        Args:
        ----
            engine (``gravitorch.engines.BaseEngine`` or ``None``,
                optional): Specifies an engine. Default: ``None``

        Returns:
        -------
            ``torchdata.dataloader2.DataLoader2``: The created dataloader.
        """


def is_dataloader2_creator_config(config: dict) -> bool:
    r"""Indicate if the input configuration is a configuration for a
    ``BaseDataLoader2Creator``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
    ----
        config (dict): Specifies the configuration to check.

    Returns:
    -------
        bool: ``True`` if the input configuration is a configuration
            for a ``BaseDataLoader2Creator`` object.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.creators.dataloader2 import is_dataloader2_creator_config
        >>> is_dataloader2_creator_config(
        ...     {"_target_": "gravitorch.creators.dataloader2.DataLoader2Creator"}
        ... )
        True
    """
    return is_object_config(config, BaseDataLoader2Creator)


def setup_dataloader2_creator(
    creator: BaseDataLoader2Creator[T] | dict,
) -> BaseDataLoader2Creator[T]:
    r"""Sets up the dataloader creator.

    The dataloader creator is instantiated from its configuration by
    using the ``BaseDataLoader2Creator`` factory function.

    Args:
    ----
        creator (``BaseDataLoader2Creator`` or dict): Specifies the
            dataloader creator or its configuration.

    Returns:
    -------
        ``BaseDataLoader2Creator``: The instantiated dataloader creator.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.creators.dataloader2 import setup_dataloader2_creator
        >>> creator = setup_dataloader2_creator(
        ...     {
        ...         "_target_": "gravitorch.creators.dataloader2.VanillaDataLoader2Creator",
        ...         "datapipe": {
        ...             "_target_": "torch.utils.data.datapipes.iter.IterableWrapper",
        ...             "iterable": [1, 2, 3, 4, 5],
        ...         },
        ...     }
        ... )
        >>> creator
        VanillaDataLoader2Creator(
          (datapipe): DataPipeCreator(
              cache=False
              datapipe={'_target_': 'torch.utils.data.datapipes.iter.IterableWrapper', 'iterable': [1, 2, 3, 4, 5]}
              deepcopy=False
            )
          (datapipe_adapter_fn): None
          (reading_service): None
        )
    """
    if isinstance(creator, dict):
        logger.info(
            "Initializing the dataloader creator from its configuration... "
            f"{str_target_object(creator)}"
        )
        creator = BaseDataLoader2Creator.factory(**creator)
    return creator



from __future__ import annotations

__all__ = ["BaseDataSourceCreator", "is_datasource_creator_config", "setup_datasource_creator"]

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from objectory import AbstractFactory
from objectory.utils import is_object_config

from gravitorch.utils.format import str_target_object

if TYPE_CHECKING:
    from gravitorch.datasources import (
        BaseDataSource,  # TODO: incorrect because used in setup
    )
    from gravitorch.engines import BaseEngine

logger = logging.getLogger(__name__)


class BaseDataSourceCreator(ABC, metaclass=AbstractFactory):
    r"""Defines the base class to create a datasource.

    Note that it is not the unique approach to create a datasource. Feel
    free to use other approaches if this approach does not fit your
    needs.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.testing import create_dummy_engine
        >>> from gravitorch.creators.datasource import DataSourceCreator
        >>> creator = DataSourceCreator({"_target_": "gravitorch.testing.DummyDataSource"})
        >>> creator
        DataSourceCreator(
          (config): {'_target_': 'gravitorch.testing.DummyDataSource'}
          (attach_to_engine): True
          (add_module_to_engine): True
        )
        >>> engine = create_dummy_engine()
        >>> datasource = creator.create(engine)
        >>> datasource
        DummyDataSource(
          (datasets):
            (train): DummyDataset(num_examples=4, feature_size=4)
            (eval): DummyDataset(num_examples=4, feature_size=4)
          (dataloader_creators):
            (train): DataLoaderCreator(
                (seed): 0
                (batch_size): 1
                (shuffle): False
              )
            (eval): DataLoaderCreator(
                (seed): 0
                (batch_size): 1
                (shuffle): False
              )
        )
    """

    @abstractmethod
    def create(self, engine: BaseEngine) -> BaseDataSource:
        r"""Creates a datasource object.

        This method is responsible to register the event handlers
        associated to the datasource.

        Args:
        ----
            engine (``gravitorch.engines.BaseEngine``): Specifies an
                engine.

        Returns:
        -------
            ``gravitorch.datasources.BaseDataSource``: The created data
                source.

        Example usage:

        .. code-block:: pycon

            >>> from gravitorch.testing import create_dummy_engine
            >>> from gravitorch.creators.datasource import DataSourceCreator
            >>> creator = DataSourceCreator({"_target_": "gravitorch.testing.DummyDataSource"})
            >>> engine = create_dummy_engine()
            >>> datasource = creator.create(engine)
            >>> datasource
            DummyDataSource(
              (datasets):
                (train): DummyDataset(num_examples=4, feature_size=4)
                (eval): DummyDataset(num_examples=4, feature_size=4)
              (dataloader_creators):
                (train): DataLoaderCreator(
                    (seed): 0
                    (batch_size): 1
                    (shuffle): False
                  )
                (eval): DataLoaderCreator(
                    (seed): 0
                    (batch_size): 1
                    (shuffle): False
                  )
            )
        """


def is_datasource_creator_config(config: dict) -> bool:
    r"""Indicates if the input configuration is a configuration for a
    ``BaseDataSourceCreator``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
    ----
        config (dict): Specifies the configuration to check.

    Returns:
    -------
        bool: ``True`` if the input configuration is a configuration
            for a ``BaseDataSourceCreator`` object.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.creators.datasource import is_datasource_creator_config
        >>> is_datasource_creator_config(
        ...     {
        ...         "_target_": "gravitorch.creators.datasource.DataSourceCreator",
        ...         "config": {"_target_": "gravitorch.testing.DummyDataSource"},
        ...     }
        ... )
        True
    """
    return is_object_config(config, BaseDataSourceCreator)


def setup_datasource_creator(creator: BaseDataSourceCreator | dict) -> BaseDataSourceCreator:
    r"""Sets up the datasource creator.

    The datasource creator is instantiated from its configuration by
    using the ``BaseDataSourceCreator`` factory function.

    Args:
    ----
        creator (``BaseDataSourceCreator`` or dict): Specifies the
            datasource creator or its configuration.

    Returns:
    -------
        ``BaseDataSourceCreator``: The instantiated datasource
            creator.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.creators.datasource import setup_datasource_creator
        >>> creator = setup_datasource_creator(
        ...     {
        ...         "_target_": "gravitorch.creators.datasource.DataSourceCreator",
        ...         "config": {"_target_": "gravitorch.testing.DummyDataSource"},
        ...     }
        ... )
        >>> creator
        DataSourceCreator(
          (config): {'_target_': 'gravitorch.testing.DummyDataSource'}
          (attach_to_engine): True
          (add_module_to_engine): True
        )
    """
    if isinstance(creator, dict):
        logger.info(
            "Initializing the datasource creator from its configuration... "
            f"{str_target_object(creator)}"
        )
        creator = BaseDataSourceCreator.factory(**creator)
    if not isinstance(creator, BaseDataSourceCreator):
        logger.warning(f"creator is not a `BaseDataSourceCreator` (received: {type(creator)})")
    return creator



r"""This module defines the base class for the data loader creators."""

from __future__ import annotations

__all__ = ["BaseDataLoaderCreator"]

from abc import ABC, abstractmethod
from collections.abc import Iterable
from typing import TYPE_CHECKING, Generic, TypeVar

from objectory import AbstractFactory
from torch.utils.data import Dataset

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine

T = TypeVar("T")


class BaseDataLoaderCreator(Generic[T], ABC, metaclass=AbstractFactory):
    r"""Defines the base class to create data loader.

    This class and its child classes are designed to be used in a data
    source.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.creators.dataloader import DataLoaderCreator
        >>> from gravitorch.testing import DummyDataset
        >>> creator = DataLoaderCreator()
        >>> creator
        DataLoaderCreator(
          (seed): 0
        )
        >>> dataset = DummyDataset()
        >>> dataloader = creator.create(dataset)
        >>> dataloader  # doctest:+ELLIPSIS
        <torch.utils.data.dataloader.DataLoader object at 0x...>
    """

    @abstractmethod
    def create(self, dataset: Dataset, engine: BaseEngine | None = None) -> Iterable[T]:
        r"""Creates a data loader given a dataset and an engine.

        The engine can be used to get the epoch, or other information
        about the training/evaluation.

        Args:
        ----
            dataset (``torch.utils.data.Dataset``): Specifies the
                dataset.
            engine (``gravitorch.engines.BaseEngine`` or ``None``,
                optional): Specifies an engine. Default: ``None``

        Returns:
        -------
            ``Iterable``: The instantiated data loader.

        Example usage:

        .. code-block:: pycon

            >>> from gravitorch.creators.dataloader import DataLoaderCreator
            >>> from gravitorch.testing import DummyDataset
            >>> creator = DataLoaderCreator()
            >>> dataset = DummyDataset()
            >>> dataloader = creator.create(dataset)
            >>> dataloader  # doctest:+ELLIPSIS
            <torch.utils.data.dataloader.DataLoader object at 0x...>
        """



__all__ = [
    "BaseDataPipeCreator",
    "ChainedDataPipeCreator",
    "DataPipeCreator",
    "DictBatcherIterDataPipeCreator",
    "EpochRandomDataPipeCreator",
    "SequentialDataPipeCreator",
    "is_datapipe_creator_config",
    "setup_datapipe_creator",
]

from gravitorch.creators.datapipe.base import (
    BaseDataPipeCreator,
    is_datapipe_creator_config,
    setup_datapipe_creator,
)
from gravitorch.creators.datapipe.chained import ChainedDataPipeCreator
from gravitorch.creators.datapipe.dictbatcher import DictBatcherIterDataPipeCreator
from gravitorch.creators.datapipe.random import EpochRandomDataPipeCreator
from gravitorch.creators.datapipe.sequential import SequentialDataPipeCreator
from gravitorch.creators.datapipe.vanilla import DataPipeCreator



from __future__ import annotations

__all__ = ["CategoricalCrossEntropy"]

import logging

from torch import Tensor
from torch.nn.functional import cross_entropy

from gravitorch.models.metrics.base_epoch import BaseStateEpochMetric
from gravitorch.models.metrics.state import BaseState, MeanErrorState

logger = logging.getLogger(__name__)


class CategoricalCrossEntropy(BaseStateEpochMetric):
    r"""Implements a metric to compute the categorical cross-entropy.

    Args:
    ----
        mode (str): Specifies the mode.
        name (str, optional): Specifies the name used to log the
            metric. Default: ``'cat_acc'``
        state (``BaseState`` or dict or ``None``, optional): Specifies
            the metric state or its configuration. If ``None``,
            ``MeanErrorState`` is instantiated. Default: ``None``

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from gravitorch.models.metrics import CategoricalCrossEntropy
        >>> metric = CategoricalCrossEntropy("eval")
        >>> metric
        CategoricalCrossEntropy(
          (mode): eval
          (name): cat_ce
          (state): MeanErrorState(num_predictions=0)
        )
        >>> metric(torch.eye(4), torch.arange(4))
        >>> metric.value()  # doctest:+ELLIPSIS
        {'eval/cat_ce_mean': 0.743668..., 'eval/cat_ce_num_predictions': 4}
        >>> metric(torch.ones(2, 3), torch.ones(2))
        >>> metric.value()  # doctest:+ELLIPSIS
        {'eval/cat_ce_mean': 0.861983..., 'eval/cat_ce_num_predictions': 6}
        >>> metric.reset()
        >>> metric(torch.ones(2, 3), torch.ones(2))
        >>> metric.value()  # doctest:+ELLIPSIS
        {'eval/cat_ce_mean': 1.098612..., 'eval/cat_ce_num_predictions': 2}
    """

    def __init__(
        self,
        mode: str,
        name: str = "cat_ce",
        state: BaseState | dict | None = None,
    ) -> None:
        super().__init__(mode=mode, name=name, state=state or MeanErrorState())

    def forward(self, prediction: Tensor, target: Tensor) -> None:
        r"""Updates the metric given a mini-batch of examples.

        Args:
        ----
            prediction (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn, num_classes)`` and type float):
                Specifies the predictions.
            target (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn)`` or ``(d0, d1, ..., dn, 1)`` and
                type long or float): Specifies the categorical
                targets. The values have to be in
                ``{0, 1, ..., num_classes-1}``.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from gravitorch.models.metrics import CategoricalCrossEntropy
            >>> metric = CategoricalCrossEntropy("eval")
            >>> metric(torch.eye(4), torch.arange(4))
            >>> metric.value()  # doctest:+ELLIPSIS
            {'eval/cat_ce_mean': 0.743668..., 'eval/cat_ce_num_predictions': 4}
        """
        self._state.update(
            cross_entropy(
                prediction.flatten(start_dim=0, end_dim=-2),
                target.flatten().long(),
                reduction="none",
            )
        )



from __future__ import annotations

__all__ = ["is_lr_scheduler_creator_config", "setup_lr_scheduler_creator"]

import logging

from objectory.utils import is_object_config

from gravitorch.creators.lr_scheduler.base import BaseLRSchedulerCreator
from gravitorch.creators.lr_scheduler.vanilla import LRSchedulerCreator
from gravitorch.utils.format import str_target_object

logger = logging.getLogger(__name__)


def is_lr_scheduler_creator_config(config: dict) -> bool:
    r"""Indicates if the input configuration is a configuration for a
    ``BaseLRSchedulerCreator``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
    ----
        config (dict): Specifies the configuration to check.

    Returns:
    -------
        bool: ``True`` if the input configuration is a configuration
            for a ``BaseLRSchedulerCreator`` object.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.creators.lr_scheduler import is_lr_scheduler_creator_config
        >>> is_lr_scheduler_creator_config(
        ...     {
        ...         "_target_": "gravitorch.creators.lr_scheduler.LRSchedulerCreator",
        ...         "lr_scheduler_config": {
        ...             "_target_": "torch.optim.lr_scheduler.StepLR",
        ...             "step_size": 5,
        ...         },
        ...     }
        ... )
        True
    """
    return is_object_config(config, BaseLRSchedulerCreator)


def setup_lr_scheduler_creator(
    creator: BaseLRSchedulerCreator | dict | None,
) -> BaseLRSchedulerCreator:
    r"""Sets up the LR scheduler creator.

    The LR scheduler creator is instantiated from its configuration by
    using the ``BaseLRSchedulerCreator`` factory function.

    Args:
    ----
        creator (``BaseLRSchedulerCreator`` or dict or ``None``):
            Specifies the LR scheduler creator or its configuration.

    Returns:
    -------
        ``BaseLRSchedulerCreator``: The instantiated LR scheduler
            creator.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from gravitorch.creators.lr_scheduler import setup_lr_scheduler_creator
        >>> creator = setup_lr_scheduler_creator(
        ...     {
        ...         "_target_": "gravitorch.creators.lr_scheduler.LRSchedulerCreator",
        ...         "lr_scheduler_config": {
        ...             "_target_": "torch.optim.lr_scheduler.StepLR",
        ...             "step_size": 5,
        ...         },
        ...     }
        ... )
        >>> creator
        LRSchedulerCreator(
          (lr_scheduler_config): {'_target_': 'torch.optim.lr_scheduler.StepLR', 'step_size': 5}
          (lr_scheduler_handler): None
          (add_module_to_engine): True
        )
    """
    if creator is None:
        creator = LRSchedulerCreator()
    if isinstance(creator, dict):
        logger.info(
            "Initializing the LR scheduler creator from its configuration... "
            f"{str_target_object(creator)}"
        )
        creator = BaseLRSchedulerCreator.factory(**creator)
    if not isinstance(creator, BaseLRSchedulerCreator):
        logger.warning(f"creator is not a `BaseLRSchedulerCreator` (received: {type(creator)})")
    return creator



from __future__ import annotations

__all__ = ["VanillaDataSource"]

import logging
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar

from coola.utils import str_indent, str_mapping

from gravitorch.creators.datastream.base import setup_datastream_creator
from gravitorch.datasources.base import BaseDataSource, IterableNotFoundError
from gravitorch.datastreams.base import BaseDataStream
from gravitorch.utils.asset import AssetManager

if TYPE_CHECKING:
    from gravitorch.creators.datastream import BaseDataStreamCreator
    from gravitorch.engines import BaseEngine

logger = logging.getLogger(__name__)

T = TypeVar("T")


class VanillaDataSource(BaseDataSource):
    r"""Implement a simple data source using datastream creators.

    Args:
    ----
        datastream_creators (``Mapping``): Specifies the datastreams
            creators or their configuration.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.datasources import VanillaDataSource
        >>> from gravitorch.creators.datastream import IterableDataStreamCreator
        >>> datasource = VanillaDataSource(
        ...     {
        ...         "train": {
        ...             "_target_": "gravitorch.creators.datastream.IterableDataStreamCreator",
        ...             "iterable": [1, 2, 3, 4],
        ...         },
        ...         "eval": IterableDataStreamCreator(["a", "b", "c"]),
        ...     }
        ... )
        >>> datasource
        VanillaDataSource(
          (train): IterableDataStreamCreator(cache=False, length=4)
          (eval): IterableDataStreamCreator(cache=False, length=3)
        )
    """

    def __init__(self, datastream_creators: Mapping[str, BaseDataStreamCreator[T] | dict]) -> None:
        self._asset_manager = AssetManager()
        self._datastream_creators: Mapping[str, BaseDataStreamCreator] = {
            key: setup_datastream_creator(value) for key, value in datastream_creators.items()
        }

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  {str_indent(str_mapping(self._datastream_creators))}\n)"
        )

    def attach(self, engine: BaseEngine) -> None:
        logger.info("Attach the data source to an engine")

    def get_asset(self, asset_id: str) -> Any:
        return self._asset_manager.get_asset(asset_id)

    def has_asset(self, asset_id: str) -> bool:
        return self._asset_manager.has_asset(asset_id)

    def get_iterable(self, iter_id: str, engine: BaseEngine | None = None) -> BaseDataStream[T]:
        if not self.has_iterable(iter_id):
            raise IterableNotFoundError(f"{iter_id} does not exist")
        return self._datastream_creators[iter_id].create(engine=engine)

    def has_iterable(self, iter_id: str) -> bool:
        return iter_id in self._datastream_creators



from __future__ import annotations

__all__ = ["CompiledModelCreator"]

import logging
from typing import TYPE_CHECKING

import torch
from coola.utils import str_indent, str_mapping
from torch.nn import Module

from gravitorch.creators.model.base import BaseModelCreator, setup_model_creator
from gravitorch.utils.format import str_pretty_json

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine

logger = logging.getLogger(__name__)


class CompiledModelCreator(BaseModelCreator):
    r"""Implements a model creator that compiles a model with
    ``torch.compile``.

    Args:
    ----
        model_creator (``BaseModelCreator`` or dict): Specifies a
            model creator or its configuration. The created model
            should be compatible with ``DistributedDataParallel``.
        compile_kwargs (dict or ``None``): Specifies some keyword
            arguments used to compile the model. Please read the
            documentation of ``torch.compile`` to see the possible
            options. Default: ``None``

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.testing import create_dummy_engine
        >>> from gravitorch.creators.model import (
        ...     CompiledModelCreator,
        ...     ModelCreator,
        ... )
        >>> creator = CompiledModelCreator(
        ...     ModelCreator({"_target_": "gravitorch.testing.DummyClassificationModel"})
        ... )
        >>> creator
        CompiledModelCreator(
          (model_creator): ModelCreator(
              (model_config): {'_target_': 'gravitorch.testing.DummyClassificationModel'}
              (attach_model_to_engine): True
              (add_module_to_engine): True
              (device_placement): AutoDevicePlacement(device=cpu)
            )
          (compile_kwargs): {}
        )
        >>> engine = create_dummy_engine()
        >>> model = creator.create(engine)
        >>> model
        OptimizedModule(
          (_orig_mod): DummyClassificationModel(
            (linear): Linear(in_features=4, out_features=3, bias=True)
            (criterion): CrossEntropyLoss()
          )
        )
    """

    def __init__(
        self,
        model_creator: BaseModelCreator | dict,
        compile_kwargs: dict | None = None,
    ) -> None:
        self._model_creator = setup_model_creator(model_creator)
        self._compile_kwargs = compile_kwargs or {}

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping(
                {
                    "model_creator": self._model_creator,
                    "compile_kwargs": str_pretty_json(self._compile_kwargs),
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def create(self, engine: BaseEngine) -> Module:
        return torch.compile(self._model_creator.create(engine), **self._compile_kwargs)
